#!/usr/bin/env python3
"""
Field Class module.

Generic, identifier, variable, category, and array field types.
"""

import numbers
from collections import OrderedDict
from functools import reduce
from operator import add


class Field:
    """Parent class for specific field types."""

    __slots__ = [
        "field_id",
        "values",
        "keys",
        "meta",
        "type",
        "headers",
        "category_slot",
        "_subset",
    ]

    def __init__(self, field_id, **kwargs):
        """Init Field class."""
        self.field_id = field_id
        self.type = "field"
        self.values = []
        self.keys = []
        self.meta = {}
        self._subset = False
        if "meta" not in kwargs:
            kwargs["meta"] = {}
        if "type" not in kwargs["meta"]:
            kwargs["meta"].update({"type": self.type})
        self.update_data(**kwargs)

    def update_data(self, **kwargs):
        """Update values and keys for an existing field."""
        for key, value in kwargs.items():
            if key != "parent":
                setattr(self, key, value)

    def update_values(self, values):
        """Update values and keys for an existing field."""
        if len(values) == len(self.values):
            self.values = values[:]

    def update_keys(self, keys):
        """Update values and keys for an existing field."""
        self.keys = keys[:]

    def get_values_by_indices(self, indices):
        """Get values for all records at matching indices."""
        values = []
        if not isinstance(indices, list):
            indices = [indices]
        for index in indices:
            try:
                values.append(self.values[index])
            except TypeError:
                pass
        return values

    def get_indices_by_values(self, values):
        """Get indices for all records matching a value."""
        if not isinstance(values, set):
            if not isinstance(values, list):
                values = [values]
            else:
                values = set(values)
        indices = [i for i, x in enumerate(self.values) if x in values]
        return indices

    def select_records(self, indices=False):
        """Create a subset of values from a list of indices."""
        values = []
        if not isinstance(indices, list):
            values = False
        else:
            values = self.get_values_by_indices(indices)
        self.subset = values
        return values

    def values_to_dict(self):
        """Create a dict of values (with keys if applicable)."""
        data = {"values": self.values}
        for key in ("keys", "category_slot", "headers"):
            if key in self.__slots__:
                if hasattr(self, key):
                    data[key] = getattr(self, key)
        return data

    @property
    def subset(self):
        """Subset property."""
        subset = []
        if isinstance(self._subset, list):
            subset = self._subset
        else:
            subset = self.values
        return subset

    @subset.setter
    def subset(self, values=False):
        """Set subset using indices."""
        if isinstance(values, list):
            self._subset = values
        else:
            self._subset = False

    @staticmethod
    def _collapse_values(values, keys=None):
        """
        Replace values with indices in list of keys.

        >>> Field._collapse_values(['A', 'A', 'B', 'A'])
        (['A', 'B'], [0, 0, 1, 0])
        """
        if keys is None:
            keys = []
        counter = len(keys)
        keys = OrderedDict(
            [(key, index) for index, key in enumerate(keys) if key is not None]
        )
        indexed_values = []
        for value in values:
            if value in keys:
                indexed_values.append(keys[value])
            elif value is not None:
                indexed_values.append(counter)
                keys[value] = counter
                counter += 1
            else:
                indexed_values.append("")
        return list(keys.keys()), indexed_values

    @staticmethod
    def _expand_values(keys, indexed_values):
        """
        Restore full values from key indices.

        >>> Field._expand_values(['A', 'B'], [0, 0, 1, 0])
        ['A', 'A', 'B', 'A']
        """
        values = [keys[index] for index in indexed_values]
        return values


class Identifier(Field):
    """Class for record identifiers."""

    def __init__(self, field_id, **kwargs):
        """Init Identifier class."""
        kwargs["type"] = "identifier"
        if "meta" not in kwargs:
            kwargs["meta"] = {}
        kwargs["meta"]["type"] = kwargs["type"]
        super().__init__(field_id, **kwargs)

    def to_set(self):
        """Create a set of identifiers."""
        unique = set()
        for value in self.values:
            unique.add(value)
        return unique

    @staticmethod
    def check_unique(entries):
        """Check all entries are unique."""
        unique = set()
        for entry in entries:
            unique.add(entry)
        return len(unique) == len(entries)

    def validate_list(self, names):
        """Ensure all list entries are unique and match identifiers."""
        valid = True
        if self.check_unique(names):
            if len(self.get_indices_by_values(names)) != len(names):
                valid = False
        else:
            valid = False
        return valid


class Variable(Field):
    """Class for variable field type."""

    def __init__(self, field_id, **kwargs):
        """Init Variable class."""
        kwargs["type"] = "variable"
        if "meta" not in kwargs:
            kwargs["meta"] = {}
        kwargs["meta"]["type"] = kwargs["type"]
        super().__init__(field_id, **kwargs)

    def get_indices_in_range(self, min_max, invert=False):
        """Get indices for all records in range."""
        if not isinstance(min_max, list):
            return -1
        if not len(min_max) == 2:
            return -2
        if not all(isinstance(x, numbers.Real) for x in min_max):
            return -3
        indices = []
        for index, value in enumerate(self.values):
            if invert:
                if value < min_max[0] or value > min_max[1]:
                    indices.append(index)
            else:
                if min_max[0] <= value <= min_max[1]:
                    indices.append(index)
        return indices

    def sum_values(self):
        """Sum values in subset."""
        try:
            total = reduce(add, self.subset)
        except TypeError:
            total = 0
        return total


class Array(Field):
    """Class for Array field type."""

    def __init__(self, field_id, **kwargs):
        """Init Array class."""
        self.category_slot = None
        if "meta" not in kwargs:
            kwargs["meta"] = {}
        if "category_slot" in kwargs:
            self.category_slot = kwargs["category_slot"]
        if self.category_slot is not None and not kwargs.get("keys"):
            slot = kwargs["category_slot"]
            cat_values = [value[slot] for value in kwargs["values"]]
            keys = kwargs.get("fixed_keys", [])
            kwargs["keys"], values = self._collapse_values(cat_values, keys)
            for index, value in enumerate(kwargs["values"]):
                value[slot] = values[index]
            kwargs["meta"].update({"category_slot": kwargs["category_slot"]})
        kwargs["meta"].update({"headers": kwargs["headers"]})
        kwargs["type"] = "array"
        kwargs["meta"]["type"] = kwargs["type"]
        super().__init__(field_id, **kwargs)

    def get_values_by_indices_for_slots(self, indices, slots):
        """Get values from specified array slots at matching indices."""
        raw_values = self.get_values_by_indices(indices)
        values = []
        if isinstance(slots, list):
            values = list(map(lambda x: [x[slot] for slot in slots], raw_values))
        else:
            values = list(map(lambda x: x[slots], raw_values))
        return values

    def update_slots(self, values, slot=0, keys=None):
        """Update values in specified slot."""
        if keys is None:
            keys = []
        if len(values) == len(self.values):
            for index, value in enumerate(values):
                self.values[index][slot] = value

    def expand_values(self):
        """Return list of full values."""
        slot = self.category_slot
        keys = self.keys
        values = []
        if slot is not None:
            for record in self.values:
                record_values = [i for i in record]
                record_values[slot] = keys[record_values[slot]]
                values.append(record_values)
        else:
            values = self.values
        return values


class MultiArray(Field):
    """Class for MultiArray field type."""

    def __init__(self, field_id, **kwargs):
        """Init MultiArray class."""
        self.category_slot = None
        kwargs["type"] = "multiarray"
        if "meta" not in kwargs:
            kwargs["meta"] = {}
        kwargs["meta"]["type"] = kwargs["type"]
        if "category_slot" in kwargs:
            self.category_slot = kwargs["category_slot"]
        if self.category_slot is not None and not kwargs.get("keys"):
            slot = kwargs["category_slot"]
            cat_values = []
            for record in kwargs["values"]:
                if record:
                    for arr in record:
                        if arr:
                            cat_values.append(arr[slot])
            keys = kwargs.get("fixed_keys", [])
            kwargs["keys"], values = self._collapse_values(cat_values, keys)
            index = 0
            keys = {k: i for i, k in enumerate(kwargs["keys"])}
            for record in kwargs["values"]:
                if record:
                    for arr in record:
                        if arr[slot] is not None:
                            arr[slot] = keys[arr[slot]]
                            index += 1
            kwargs["meta"].update({"category_slot": kwargs["category_slot"]})
        kwargs["meta"].update({"headers": kwargs["headers"]})
        super().__init__(field_id, **kwargs)

    def expand_values(self):
        """Return list of full values."""
        slot = self.category_slot
        keys = self.keys
        values = []
        if slot is not None:
            for record in self.values:
                record_values = []
                for arr in record:
                    arr_values = [i for i in arr]
                    arr_values[slot] = keys[arr_values[slot]]
                    record_values.append(arr_values)
                values.append(record_values)
        else:
            values = self.values
        return values


class Category(Field):
    """Class for Category field type."""

    def __init__(self, field_id, **kwargs):
        """Init Category class."""
        kwargs["type"] = "category"
        if "meta" not in kwargs:
            kwargs["meta"] = {}
        kwargs["meta"]["type"] = kwargs["type"]
        if "keys" not in kwargs or kwargs["keys"] is None:
            keys = kwargs.get("fixed_keys", [])
            kwargs["keys"], kwargs["values"] = self._collapse_values(
                kwargs["values"], keys
            )
        super().__init__(field_id, **kwargs)

    def expand_values(self):
        """Return list of full values."""
        return self._expand_values(self.keys, self.values)

    def second_func(self):
        """Do something else."""
        return self.type


__all__ = ["Array", "Category", "Field", "Identifier", "MultiArray", "Variable"]


if __name__ == "__main__":
    import doctest

    doctest.testmod()



#!/usr/bin/env python3

"""
BlobTools2 - assembly exploration, QC and filtering.

usage: blobtools [<command>] [<args>...] [-h|--help] [--version]

commands:
    add             add data to a BlobDir
    create          create a new BlobDir
    filter          filter a BlobDir
    host            host interactive view of all BlobDirs in a directory
    replace         call blobtools add with --replace flag
    remove          remove one or more fields from a BlobDir
    validate        validate a BlobDir
    view            generate plots using BlobToolKit Viewer
    -h, --help      show this
    -v, --version   show version number
See 'blobtools <command> --help' for more information on a specific command.

examples:
    # 1. Create a new BlobDir from a FASTA file:
    ./blobtools create --fasta examples/assembly.fasta BlobDir

    # 2. Create a new BlobDir from a BlobDB:
    ./blobtools create --blobdb examples/blobDB.json BlobDir

    # 3. Add Coverage data from a BAM file:
    ./blobtools add --cov examples/assembly.reads.bam BlobDir

    # 4. Assign taxonomy from BLAST hits:
    ./blobtools add add --hits examples/blast.out --taxdump ../taxdump BlobDir

    # 5. Add BUSCO results:
    ./blobtools add --busco examples/busco.tsv BlobDir

    # 6. Host an interactive viewer:
    ./blobtools host BlobDir

    # 7. Filter a BlobDir:
    ./blobtools filter --param length--Min=5000 --output BlobDir_len_gt_5000 BlobDir
"""


import os
import sys

from docopt import DocoptExit
from docopt import docopt
from pkg_resources import working_set
from tolkein import tolog

from .lib.version import __version__

LOGGER = tolog.logger(__name__)


def cli():
    """Entry point."""
    args = {}
    if len(sys.argv) > 2:
        try:
            args = docopt(__doc__, help=False, version=__version__)
        except DocoptExit:
            args = {
                "<command>": " ".join(sys.argv[1:2]),
            }
    elif len(sys.argv) > 1:
        try:
            args = docopt(__doc__, help=False, version=__version__)
        except DocoptExit:
            args = {"<command>": sys.argv[1]}
    else:
        print(__doc__)
    if "<command>" in args and args["<command>"]:
        args.update({"<tool>": os.path.basename(sys.argv[0])})
        # load <command> from entry_points
        command_index = sys.argv.index(args["<command>"])
        if args["<command>"] in ["create", "replace"]:
            if len(sys.argv) > command_index + 1:
                if args["<command>"] == "replace" and "--replace" not in sys.argv:
                    sys.argv.insert(command_index + 1, "--replace")
                args["<command>"] = "add"
            sys.argv[command_index] = "add"
        for entry_point in working_set.iter_entry_points("%s.subcmd" % args["<tool>"]):
            if entry_point.name == args["<command>"]:
                subcommand = entry_point.load()
                sys.exit(subcommand())
        LOGGER.error(
            "'%s %s' is not a valid command", args["<tool>"], args["<command>"]
        )
        sys.exit(1)
    raise DocoptExit



#!/usr/bin/env python3
"""Parse BUSCO results into MultiArray Field."""

import re
from collections import defaultdict

from ..lib import file_io
from .field import MultiArray


def parse_busco(busco_file, identifiers):  # pylint: disable=too-many-locals
    """Parse BUSCO results into a MultiArray."""
    data = file_io.read_file(busco_file)
    if not data:
        print("WARNING file %s is empty" % busco_file)
        return None
    lines = data.split("\n")
    version = lines[0].split(":")[1].strip()
    desc = re.split(r":\s*|\(|\)\s*|,\s*", lines[1])
    meta = {
        "version": version,
        "set": desc[1].strip(),
        "count": int(desc[7].strip()),
        "file": busco_file,
    }
    version = int(version.split(".")[0])
    if version < 4:
        rows = [re.split("\t", line) for line in lines[5:]]
        meta["set"] = re.search(
            r"-l\s.*?\/*(\w+_odb\d+)\/", lines[2].split(":")[1].strip()
        )[1]
        columns = re.split(r"# |\t", lines[4])[1:]
        try:
            contig_index = columns.index("Contig")
        except ValueError:
            contig_index = columns.index("Sequence")
    else:
        rows = [re.split("\t", line) for line in lines[3:]]
        columns = re.split(r"# |\t", lines[2])[1:]
        contig_index = columns.index("Sequence")
    meta["field_id"] = "%s_busco" % meta["set"]
    busco_index = columns.index("Busco id")
    status_index = columns.index("Status")
    results = defaultdict(list)
    for row in rows:
        if len(row) > contig_index:
            if version < 4:
                contig = row[contig_index]
            else:
                contig = row[contig_index].split(":")[0]
            results[contig].append([row[busco_index], row[status_index]])
    if not identifiers.validate_list(list(results.keys())):
        # try removing _\d+ suffix added by prokka-based busco
        res = {}
        for contig, values in results.items():
            ctg = re.sub(r"_\d+$", "", contig)
            if ctg not in res:
                res[ctg] = values
            else:
                res[ctg] += values
        results = res
        if not identifiers.validate_list(list(results.keys())):
            raise UserWarning(
                "Contig names in the Busco file did not match dataset identifiers."
            )
    values = [results[id] if id in results else [] for id in identifiers.values]
    busco_field = MultiArray(
        meta["field_id"],
        values=values,
        meta=meta,
        headers=("Busco id", "Status"),
        parents=["children"],
        category_slot=1,
    )
    return busco_field


def parse(files, **kwargs):
    """Parse all BUSCO files."""
    parsed = []
    for file in files:
        busco = parse_busco(file, identifiers=kwargs["dependencies"]["identifiers"])
        if busco is not None:
            parsed.append(busco)
    return parsed


def parent():
    """Set standard metadata for BUSCO."""
    busco = {"datatype": "mixed", "type": "array", "id": "busco", "name": "Busco"}
    return [busco]


def busco_score(values, total):
    """Calculate BUSCO score."""
    fragmented = set()
    buscos = defaultdict(int)
    for contig in values:
        for busco in contig:
            if busco[1] == "Fragmented":
                fragmented.add(busco[0])
            buscos[busco[0]] += 1
    present = len(buscos.keys())
    complete = present - len(fragmented)
    duplicated = len([value for value in buscos.values() if value > 1])
    scores = {
        "c": complete,
        "d": duplicated,
        "m": total - present,
        "f": len(fragmented),
        "t": total,
        "s": complete - duplicated,
    }
    string = "C:{:.1%}".format(scores["c"] / total)
    string += "[S:{:.1%},".format(scores["s"] / total)
    string += "D:{:.1%}],".format(scores["d"] / total)
    string += "F:{:.1%},".format(scores["f"] / total)
    string += "M:{:.1%},".format(scores["m"] / total)
    string += "n:{:d}".format(total)
    scores.update({"string": string})
    return scores


def summarise(indices, fields, **kwargs):  # pylint: disable=unused-argument
    """Summarise BUSCOs."""
    summary = {}
    for lineage in fields["lineages"]:
        values = fields[lineage].expand_values()
        values = [values[i] for i in indices]
        total = fields[lineage].meta["count"]
        lineage = lineage.replace("_busco", "")
        summary.update({lineage: busco_score(values, total)})
    return summary


def remove_from_meta(meta):
    """Delete all BUSCO fields."""
    field_ids = []
    if meta.has_field("busco"):
        field_ids = meta.remove_field("busco")
    return field_ids



#!/usr/bin/env python3
"""
Run BlobToolKit Pipeline.

Usage: btk pipeline run [--dry-run] [--tool STRING] [--unlock]
                        [--workdir DIR] [--threads INT] --config YAML

Options:
    --config YAML  YAML format configuration filename.
    --dry-run      Dry run flag.
    --threads INT  Number of threads to use. [Default: 32]
    --tool STRING  Pipeline tool to run. [Default: blobtoolkit]
    --unlock       Flag to unlock working directory.
    --workdir DIR  Full Path of working directory.
                   Default is the directory containing config file.
"""

import logging
import os
import shlex
import shutil
import signal
import subprocess

from docopt import DocoptExit
from docopt import docopt

logger_config = {
    "level": logging.INFO,
    "format": "%(asctime)s [%(levelname)s] line %(lineno)d %(message)s",
    "filemode": "w",
}
logging.basicConfig(**logger_config)
logger = logging.getLogger()


def run_command(cmd):
    """Run a command and capture CTRL-C."""
    script_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
    p = subprocess.Popen(shlex.split(cmd), stdin=subprocess.PIPE, cwd=script_dir)
    exit_code = p.wait()
    return exit_code


def unlock_working_directory(workdir):
    """Unlock Snakemake working directory after a failed run."""
    script_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
    snakefile = os.path.join(script_dir, "blobtoolkit.smk")
    cmd = """
    snakemake -p \
          -j 1 \
          --directory %s/blobtoolkit \
          --configfile %s/config.yaml \
          --unlock \
          -s %s
    """ % (
        workdir,
        workdir,
        snakefile,
    )
    exit_code = run_command(cmd)


def run_pipeline(workdir, args):
    """Run Snakemake pipeline."""
    script_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
    snakefile = os.path.join(script_dir, "%s.smk" % args["--tool"])
    if "--dry-run" in args and args["--dry-run"] is True:
        dry_run = "-n"
    else:
        dry_run = ""
    cmd = """
    snakemake -p %s \
          -j %s \
          --directory %s/blobtoolkit \
          --configfile %s/config.yaml \
          -s %s
    """ % (
        dry_run,
        args["--threads"],
        workdir,
        workdir,
        snakefile,
    )
    exit_code = run_command(cmd)


def main():
    """Entry point."""
    try:
        args = docopt(__doc__)
    except DocoptExit:
        raise DocoptExit
    try:
        config = os.path.abspath(args["--config"])
        if args["--workdir"] is not None:
            workdir = os.path.abspath(args["--workdir"])
        else:
            workdir = os.path.abspath(os.path.dirname(config))
        if config != "%s/config.yaml" % workdir:
            shutil.copy2(config, "%s/config.yaml" % workdir)
        if args["--unlock"]:
            exit_code = unlock_working_directory(workdir)
            exit(exit_code)
        exit_code = run_pipeline(workdir, args)
        exit(exit_code)

    except Exception as err:
        logger.error(err)
        exit(1)



"""BlobToolKit."""

from .blobtools import cli
from .lib import add
from .lib import filter
from .lib import host
from .lib import view

# from .field import Array, Category, Field, Identifier, MultiArray, Variable
# from .taxdump import Taxdump
# from .dataset import Metadata



#!/usr/bin/env python3
"""Add links to metadata."""

import re
from urllib.error import HTTPError, URLError
from urllib.parse import quote
from urllib.request import urlopen


def _expand_link(url, values, identifiers):
    """Expand a link url using dict of values."""
    parts = re.split(r'\{|\}', url)
    for i in range(1, len(parts), 2):
        if parts[i] == 'id':
            parts[i] = identifiers[0]
        else:
            parts[i] = str(values[parts[i]])
    return ''.join(parts)


def _test_link(url):
    """Check url exists."""
    url = quote(url, ':/')
    try:
        urlopen(url)
    except HTTPError as error:
        print('HTTPError: {}'.format(error.code))
    except URLError as error:
        print('URLError: {}'.format(error.reason))
    else:
        return True


def add(string, meta, identifiers, skip_test):
    """Add a link to meta."""
    path, url = string.split('=', 1)
    keys = path.split('.')
    links = meta.links
    values = meta
    if keys[0] == 'position':
        skip_test = True
    if skip_test:
        valid = True
    else:
        for key in keys[:-2]:
            if isinstance(values, dict):
                values = values[key]
            else:
                values = getattr(values, key)
        valid = _test_link(_expand_link(url, values, identifiers))
    if valid:
        for key in keys[:-1]:
            if key not in links:
                links[key] = {}
            links = links[key]
        links.update({keys[-1]: url})



#!/usr/bin/env python3

# pylint: disable=no-member, too-many-branches, too-many-locals, too-many-statements, too-many-nested-blocks

"""
Filter a BlobDir.

Usage:
    blobtools filter [--param STRING...] [--query-string STRING] [--json JSON]
                     [--list TXT] [--invert] [--output DIRECTORY]
                     [--fasta FASTA] [--fastq FASTQ...] [--suffix STRING]
                     [--cov BAM] [--summary FILENAME] [--summary-rank RANK]
                     [--table FILENAME] [--table-fields STRING]
                     [--taxdump DIRECTORY] [--taxrule STRING] [--text TXT] [--text-header]
                     [--text-delimiter STRING] [--text-id-column INT] DIRECTORY

Arguments:
    DIRECTORY                   Existing BlobDir dataset directory.

Options:
    --param STRING            String of type param=value.
    --query-string STRING     List of param=value pairs from url query string.
    --json JSON               JSON format list file as generated by BlobtoolKit Viewer.
    --list TXT                Space or newline separated list of identifiers.
    --invert                  Invert filter (exclude matching records).
    --output DIRECTORY        Path to directory to generate a new, filtered BlobDir.
    --fasta FASTA             FASTA format assembly file to be filtered.
    --fastq FASTQ             FASTQ format read file to be filtered (requires --cov).
    --cov BAM                 BAM/SAM/CRAM read alignment file.
    --text TXT                generic text file to be filtered.
    --text-delimiter STRING   text file delimiter. [Default: whitespace]
    --text-id-column INT      index of column containing identifiers (1-based). [Default: 1]
    --text-header             Flag to indicate first row of text file contains field names. [Default: False]
    --suffix STRING           String to be added to filtered filename. [Default: filtered]
    --summary FILENAME        Generate a JSON-format summary of the filtered dataset.
    --summary-rank RANK       Taxonomic level for summary. [Default: phylum]
    --table FILENAME          Tabular output of filtered dataset.
    --table-fields STRING     Comma separated list of field IDs to include in the
                              table output. Use 'plot' to include all plot axes.
                              [Default: plot]
    --taxdump DIRECTORY       Location of NCBI new_taxdump directory.
    --taxrule STRING          Taxrule used when processing hits.
"""

import math
import re
import sys
import urllib
from collections import defaultdict

from docopt import docopt

from ..lib import busco
from ..lib import cov
from ..lib import fasta
from ..lib import file_io
from ..lib import hits
from ..lib import taxid
from ..lib import text

# from dataset import Metadata
from .fetch import fetch_field
from .fetch import fetch_metadata

# from taxdump import Taxdump
from .field import Category
from .field import Identifier
from .field import MultiArray
from .field import Variable
from .version import __version__

FIELDS = [
    {"flag": "--fasta", "module": fasta, "depends": ["identifiers"]},
    {
        "flag": "--fastq",
        "module": cov,
        "depends": ["identifiers"],
        "requires": ["--cov"],
    },
    {
        "flag": "--text",
        "module": text,
        "depends": ["identifiers"],
        "requires": ["--text-delimiter", "--text-header", "--text-id-column"],
    },
]

SUMMARY = [
    {"title": "taxonomy", "module": taxid, "depends": []},
    {
        "title": "baseComposition",
        "module": fasta,
        "depends": ["gc", "ncount", "length"],
    },
    {"title": "hits", "module": hits, "depends": ["length", "gc"]},
    {"title": "busco", "module": busco, "depends": []},
    {"title": "readMapping", "module": cov, "depends": ["length"]},
]


def parse_params(args, meta):
    """Parse and perform sanity checks on filer parameters."""
    strings = args.get("--param", [])
    valid = {
        "variable": ["Min", "Max", "Inv"],
        "category": ["Keys", "Inv"],
        "multiarray": ["Keys", "MinLength", "MaxLength", "Inv"],
    }
    if args.get("--query-string"):
        qstr = args["--query-string"]
        qstr = re.sub(r"^.*\?", "", qstr)
        qstr = re.sub(r"#.*$", "", qstr)
        strings += urllib.parse.unquote(qstr).split("&")
    params = defaultdict(dict)
    for string in strings:
        try:
            key, value = string.split("=")
        except ValueError:
            print("WARN: Skipping string '%s', not a valid parameter" % string)
            continue
        try:
            field_id, param = key.split("--")
        except ValueError:
            print("WARN: Skipping string '%s', not a valid parameter" % string)
            continue
        if meta.has_field(field_id):
            field_meta = meta.field_meta(field_id)
            # if field_meta.get('range'):
            #     field_type = 'variable'
            # else:
            #     field_type = 'category'
            if param in valid[field_meta["type"]]:
                params[field_id].update({param: value})
            else:
                print(
                    "WARN: '%s' is not a valid parameter for field '%s'"
                    % (param, field_id)
                )
        else:
            print("WARN: Skipping field '%s', not present in dataset" % field_id)
    return dict(params)


def filter_by_params(meta, directory, indices, params, invert_all):
    """Filter included set using params."""
    all_indices = indices[0:]
    for field_id, filters in params.items():
        field = fetch_field(directory, field_id, meta)
        invert = False
        if filters.get("Inv"):
            invert = True
        if isinstance(field, Category):
            keys = field.keys
            if filters.get("Keys"):
                keys = [
                    int(x) if x.isdigit() else field.keys.index(x)
                    for x in filters["Keys"].split(",")
                ]
                if not invert:
                    keys = [i for i, x in enumerate(field.keys) if i not in keys]
                keys = set(keys)
                indices = [i for i in indices if field.values[i] in keys]
        elif isinstance(field, Variable):
            low = -math.inf
            high = math.inf
            if filters.get("Min"):
                low = float(filters["Min"])
            if filters.get("Max"):
                high = float(filters["Max"])
            if invert:
                indices = [
                    i
                    for i in indices
                    if field.values[i] < low or field.values[i] > high
                ]
            else:
                indices = [i for i in indices if low <= field.values[i] <= high]
        elif isinstance(field, MultiArray):
            low = -math.inf
            high = math.inf
            length = False
            if filters.get("MinLength"):
                low = int(filters["MinLength"])
                length = True
            if filters.get("MaxLength"):
                high = int(filters["MaxLength"])
                length = True
            if length:
                if invert:
                    indices = [
                        i
                        for i in indices
                        if len(field.values[i]) < low or len(field.values[i]) > high
                    ]
                else:
                    indices = [
                        i for i in indices if low <= len(field.values[i]) <= high
                    ]
            if filters.get("Keys"):
                keys = [
                    int(x) if x.isdigit() else field.keys.index(x)
                    for x in filters["Keys"].split(",")
                ]
                cat_i = int(field.category_slot)
                if not invert:
                    keys = [i for i, x in enumerate(field.keys) if i not in keys]
                keys = set(keys)
                new_indices = []
                for i in indices:
                    for j in field.values[i]:
                        if j[cat_i] in keys:
                            new_indices.append(i)
                            break
                indices = new_indices
    if invert_all:
        inverted = [i for i in all_indices if i not in indices]
        return inverted
    return indices


def filter_by_json(identifiers, indices, json_file, invert):
    """Filter included set using json file."""
    data = file_io.load_yaml(json_file)
    id_set = set(data["identifiers"])
    if not invert:
        indices = [i for i in indices if identifiers[i] in id_set]
    else:
        indices = [i for i in indices if identifiers[i] not in id_set]
    return indices


def create_filtered_dataset(dataset_meta, indir, outdir, indices):
    """Write filtered records to new dataset."""
    meta = dataset_meta.to_dict()
    meta.update(
        {"fields": [], "origin": dataset_meta.dataset_id, "records": len(indices)}
    )
    meta.pop("id")
    meta = fetch_metadata(outdir, meta=meta)
    # meta = fetch_metadata(outdir, **args)
    for field_id in dataset_meta.list_fields():
        field_meta = dataset_meta.field_meta(field_id)
        if not field_meta.get("children"):
            field_meta.pop("data", False)
            keys = None
            slot = None
            headers = None
            full_field = fetch_field(indir, field_id, dataset_meta)
            if isinstance(full_field, (Variable, Identifier)):
                values = [full_field.values[i] for i in indices]
                if isinstance(full_field, Variable):
                    field_meta.update({"range": [min(values), max(values)]})
                    if field_id == "length":
                        meta.assembly.update({"span": sum(values)})
                        meta.assembly.update({"scaffold-count": len(values)})
            elif isinstance(full_field, Category):
                full_values = full_field.expand_values()
                values = [full_values[i] for i in indices]
            else:
                full_values = full_field.expand_values()
                values = [full_values[i] for i in indices]
                slot = full_field.category_slot
                try:
                    headers = full_field.headers
                except AttributeError:
                    pass
                if field_meta.get("parent"):
                    parent_field = fetch_field(
                        outdir, field_meta["parent"], dataset_meta
                    )
                    if parent_field:
                        keys = parent_field.keys
            field = type(full_field)(
                field_id,
                meta=field_meta,
                values=values,
                fixed_keys=keys,
                category_slot=slot,
                headers=headers,
            )
            parents = dataset_meta.field_parent_list(field_id)
            meta.add_field(parents, **field_meta, field_id=field_id)
            json_file = "%s/%s.json" % (outdir, field.field_id)
            file_io.write_file(json_file, field.values_to_dict())
    file_io.write_file("%s/meta.json" % outdir, meta.to_dict())


def main(args):
    """Entrypoint for blobtools filter."""
    meta = fetch_metadata(args["DIRECTORY"], **args)
    params = parse_params(args, meta)
    identifiers = fetch_field(args["DIRECTORY"], "identifiers", meta)
    indices = [index for index, value in enumerate(identifiers.values)]
    invert = args["--invert"]
    if params:
        indices = filter_by_params(meta, args["DIRECTORY"], indices, params, invert)
    if args["--json"]:
        indices = filter_by_json(identifiers.values, indices, args["--json"], invert)
    if args["--output"]:
        create_filtered_dataset(meta, args["DIRECTORY"], args["--output"], indices)
    ids = [identifiers.values[i] for i in indices]
    for field in FIELDS:
        if args[field["flag"]]:
            requirements = True
            if field.get("requires"):
                for flag in field["requires"]:
                    if flag not in args:
                        print(
                            "WARN: '%s' must be set to use option '%s'"
                            % (flag, field["flag"])
                        )
                        requirements = False
            if not requirements:
                continue
            field["module"].apply_filter(ids, args[field["flag"]], **args)
    if args["--table"]:
        full_field_ids = args["--table-fields"].split(",")
        expanded_ids = ["index", "identifiers"]
        field_ids = []
        alt_ids = {field_id: field_id for field_id in expanded_ids}
        for full_id in full_field_ids:
            try:
                field_id, alt_id = full_id.split("=")
                field_ids.append(field_id)
                alt_ids[field_id] = alt_id
            except ValueError:
                field_ids.append(full_id)
                alt_ids[full_id] = full_id
        fields = {"identifiers": fetch_field(args["DIRECTORY"], "identifiers", meta)}
        for field_id in field_ids:
            if field_id == "plot":
                for axis in ["x", "z", "y", "cat"]:
                    if axis in meta.plot:
                        expanded_ids.append(meta.plot[axis])
                        alt_ids.update({meta.plot[axis]: meta.plot[axis]})
                        fields[meta.plot[axis]] = fetch_field(
                            args["DIRECTORY"], meta.plot[axis], meta
                        )
            else:
                expanded_ids.append(field_id)
                alt_ids.update({field_id: field_id})
                fields[field_id] = fetch_field(args["DIRECTORY"], field_id, meta)
        table = [[alt_ids[field_id] for field_id in expanded_ids]]
        for i in indices:
            record = []
            for field_id in expanded_ids:
                if field_id == "index":
                    record.append(i)
                else:
                    value = fields[field_id].values[i]
                    if fields[field_id].keys:
                        value = fields[field_id].keys[value]
                    record.append(value)
            table.append(record)
        file_io.write_file(args["--table"], table)
    if args["--summary"]:
        summary_stats = {}
        for section in SUMMARY:
            requirements = True
            if section.get("requires"):
                for flag in section["requires"]:
                    if not args[flag]:
                        print(
                            "WARN: '%s' must be set to generate '%s' summary"
                            % (flag, section["title"])
                        )
                        requirements = False
            if not requirements:
                continue
            fields = {}
            if section.get("depends"):
                for field in section["depends"]:
                    fields.update({field: fetch_field(args["DIRECTORY"], field, meta)})
            if section["title"] == "hits":
                taxrule = args.get("--taxrule", None)
                if taxrule is None:
                    taxrule = meta.plot.get("cat", None)
                    if taxrule is not None:
                        taxrule = re.sub(r"_[^_]+$", "", taxrule)
                        args["--taxrule"] = taxrule
                    else:
                        continue
                field = "%s_%s" % (taxrule, args["--summary-rank"])
                fields.update({"hits": fetch_field(args["DIRECTORY"], field, meta)})
                if "y" in meta.plot:
                    fields.update(
                        {"cov": fetch_field(args["DIRECTORY"], meta.plot["y"], meta)}
                    )
            if section["title"] == "busco":
                lineages = []
                for field in meta.list_fields():
                    if field.endswith("_busco"):
                        lineages.append(field)
                        fields.update(
                            {field: fetch_field(args["DIRECTORY"], field, meta)}
                        )
                fields.update({"lineages": lineages})
            if section["title"] == "readMapping":
                libraries = []
                for field in meta.list_fields():
                    if field.endswith("_cov") and not field.endswith("_read_cov"):
                        library = field.replace("_cov", "")
                        libraries.append(library)
                        fields.update(
                            {field: fetch_field(args["DIRECTORY"], field, meta)}
                        )
                fields.update({"libraries": libraries})
            summary_stats.update(
                {
                    section["title"]: section["module"].summarise(
                        indices, fields, **args, meta=meta, stats=summary_stats
                    )
                }
            )
        stats = {}
        if "hits" in summary_stats:
            nohit_span = 0
            span = summary_stats["hits"]["total"]["span"]
            if "no-hit" in summary_stats["hits"]:
                nohit_span = summary_stats["hits"]["no-hit"]["span"]
                stats.update({"noHit": float("%.3f" % (nohit_span / span))})
            else:
                stats.update({"noHit": 0})
            if "taxonomy" in summary_stats and "target" in summary_stats["taxonomy"]:
                if summary_stats["taxonomy"]["target"] in summary_stats["hits"]:
                    target_span = summary_stats["hits"][
                        summary_stats["taxonomy"]["target"]
                    ]["span"]
                    stats.update(
                        {"target": float("%.3f" % (target_span / (span - nohit_span)))}
                    )
                elif "target" in summary_stats["hits"]:
                    target_span = summary_stats["hits"]["target"]["span"]
                    stats.update(
                        {"target": float("%.3f" % (target_span / (span - nohit_span)))}
                    )
                    del summary_stats["hits"]["target"]
                else:
                    stats.update({"target": 0})
            ratio = (
                summary_stats["hits"]["total"]["span"]
                / summary_stats["hits"]["total"]["n50"]
            )
            if ratio >= 100:
                ratio = int(float("%.3g" % ratio))
            else:
                ratio = float("%.3g" % ratio)
            stats.update({"spanOverN50": ratio})
        summary_stats.update({"stats": stats})
        file_io.write_file(args["--summary"], {"summaryStats": summary_stats})


def cli():
    """Entry point."""
    if len(sys.argv) == sys.argv.index(__name__.split(".")[-1]) + 1:
        args = docopt(__doc__, argv=[])
    else:
        args = docopt(__doc__, version=__version__)
    main(args)


if __name__ == "__main__":
    cli()



#!/usr/bin/env python3

"""
BlobToolKit Pipeline.

Usage: blobtools pipeline [<command>] [<args>...] [-h|--help] [--version]

Commands:
    data                      Fetch pipeline data
    run                       Run BlobToolKit Pipeline
    add-summary-to-metadata   Pipeline helper script
    chunk-fasta               Pipeline helper script
    count-busco-genes         Pipeline helper script
    extract-busco-genes       Pipeline helper script
    generate-config           Pipeline helper script
    generate-static-images    Pipeline helper script
    transfer-completed        Pipeline helper script
    unchunk-blast             Pipeline helper script
    window-stats              Pipeline helper script
    -h, --help      Show this
    -v, --version   Show version number
See 'blobtools pipeline <command> --help' for more information on a specific command.

"""

import sys

from docopt import DocoptExit
from docopt import docopt
from pkg_resources import working_set

from .lib.version import __version__


def main():
    """Entry point."""
    if len(sys.argv) > 2:
        try:
            args = docopt(__doc__, help=False, version=__version__)
        except DocoptExit:
            args = {"<command>": sys.argv[2]}
        if args["<command>"]:
            # load <command> from entry_points
            for entry_point in working_set.iter_entry_points("pipeline.subcmd"):
                if entry_point.name == args["<command>"]:
                    subcommand = entry_point.load()
                    sys.exit(subcommand())
    print(__doc__)
    raise DocoptExit



#!/usr/bin/env python3
"""Add key values to metadata."""

import re


def add(string, meta, replace):
    """Add a key value to meta."""
    path, value = string.split('=')
    if re.match(r'^\[.+\]$', value):
        value = value.strip('[]').split(',')
    keys = path.split('.')
    if len(keys) == 1:
        setattr(meta, keys[0], value)
        return True
    if hasattr(meta, keys[0]):
        current = getattr(meta, keys[0])
    else:
        current = {}
        setattr(meta, keys[0], current)
    for key in keys[1:-1]:
        if key not in current:
            current[key] = {}
        current = current[key]
    if not replace and keys[-1] in current:
        if isinstance(current[keys[-1]], list):
            if isinstance(value, list):
                value = current[keys[-1]] + value
            else:
                val = value
                value = current[keys[-1]][:]
                value.append(val)
        elif current[keys[-1]]:
            value = [current[keys[-1]], value]
    current.update({keys[-1]: value})
    return True



#!/usr/bin/env python3
"""Convert a blobDB to BlobDir Fields."""

# pylint: disable=too-many-locals

import math
from collections import defaultdict
from pathlib import Path

from ..lib import cov
from ..lib import file_io
from ..lib import hits
from .field import Category
from .field import Identifier
from .field import MultiArray
from .field import Variable


def field_name_from_path(path):
    """Extract field name from file path."""
    parts = Path(path).stem.split(".")
    field_name = parts[-1]
    if len(parts) > 1:
        if parts[-1] in ("bam", "sam", "cram"):
            field_name = parts[-2]
    return field_name


def values_from_blob_db(blob_db):
    """Read values from a blobDB into a dict of lists of values."""
    values = defaultdict(list)
    for identifier in blob_db["order_of_blobs"]:
        blob = blob_db["dict_of_blobs"][identifier]
        values["lengths"].append(blob.get("length", 0))
        values["gcs"].append(blob.get("gc", 0))
        values["n_counts"].append(blob.get("n_count", 0))
        for cov_lib in blob_db["covLibs"].keys():
            values["%s_cov" % cov_lib].append(blob["covs"].get(cov_lib, 0))
            values["%s_read_cov" % cov_lib].append(blob["read_cov"].get(cov_lib, 0))
        for tax_rule in blob_db["taxrules"]:
            for rank, results in blob["taxonomy"][tax_rule].items():
                values["%s_%s" % (tax_rule, rank)].append(results.get("tax", "no-hit"))
                values["%s_%s_score" % (tax_rule, rank)].append(
                    float(results.get("score", 0))
                )
                values["%s_%s_cindex" % (tax_rule, rank)].append(
                    int(results.get("c_index", 0))
                )
    return values


def hits_from_blob_db(blob_db, tax_rule):
    """Read values from a blobDB into a dict of lists of values."""
    values = defaultdict(list)
    for identifier in blob_db["order_of_blobs"]:
        if identifier in values and tax_rule == "bestsumorder":
            continue
        hit_list = blob_db["dict_of_blobs"][identifier]["hits"]
        for tax_lib in blob_db["hitLibs"].keys():
            if tax_lib in hit_list:
                for detail in hit_list[tax_lib]:
                    try:
                        values[identifier].append(
                            {
                                "taxid": int(detail["taxId"]),
                                "score": int(detail["score"]),
                            }
                        )
                    except ValueError:
                        pass
    values = [values.get(identifier, []) for identifier in blob_db["order_of_blobs"]]
    return values


def parse(file, **kwargs):
    """Parse all synonym files."""
    blob_db = file_io.load_yaml(file)
    kwargs["meta"].assembly.update({"file": blob_db["assembly_f"]})
    parsed = []
    identifiers = kwargs["dependencies"]["identifiers"]
    if not identifiers:
        identifiers = Identifier(
            "identifiers",
            meta={"field_id": "identifiers"},
            values=blob_db["order_of_blobs"],
            parents=[],
        )
        kwargs["meta"].assembly.update({"scaffold-count": len(identifiers.values)})
        parsed.append(identifiers)
    values = values_from_blob_db(blob_db)
    kwargs["meta"].assembly.update({"span": sum(values["lengths"])})
    parsed.append(
        Variable(
            "gc",
            meta={
                "preload": True,
                "scale": "scaleLinear",
                "field_id": "gc",
                "name": "GC",
                "datatype": "float",
                "range": [min(values["gcs"]), max(values["gcs"])],
            },
            values=values["gcs"],
            parents=[],
        )
    )
    _min = min(values["lengths"])
    parsed.append(
        Variable(
            "length",
            meta={
                "field_id": "length",
                "preload": True,
                "scale": "scaleLog",
                "name": "Length",
                "clamp": 100 if _min == 0 else False,
                "datatype": "integer",
                "range": [_min, max(values["lengths"])],
            },
            parents=[],
            values=values["lengths"],
        )
    )
    parsed.append(
        Variable(
            "ncount",
            meta={
                "field_id": "ncount",
                "scale": "scaleLinear",
                "name": "N count",
                "datatype": "integer",
                "range": [min(values["n_counts"]), max(values["n_counts"])],
            },
            values=values["n_counts"],
            parents=[],
        )
    )
    if "z" not in kwargs["meta"].plot:
        kwargs["meta"].plot.update({"z": "length"})
    if "x" not in kwargs["meta"].plot:
        kwargs["meta"].plot.update({"x": "gc"})
    cov_range = [math.inf, -math.inf]
    read_cov_range = [math.inf, -math.inf]
    for cov_lib, cov_meta in blob_db["covLibs"].items():
        cov_file_name = field_name_from_path(blob_db["covLibs"][cov_lib]["f"])
        covs = values["%s_cov" % cov_lib]
        read_covs = values["%s_read_cov" % cov_lib]
        cov_range = [min(covs + [cov_range[0]]), max(covs + [cov_range[1]])]
        read_cov_range = [
            min(read_covs + [read_cov_range[0]]),
            max(read_covs + [read_cov_range[1]]),
        ]
        if "y" not in kwargs["meta"].plot:
            kwargs["meta"].plot.update({"y": "%s_cov" % cov_file_name})
        parsed.append(
            Variable(
                "%s_cov" % cov_file_name,
                values=covs,
                meta={"field_id": "%s_cov" % cov_file_name, "file": cov_meta["f"]},
                parents=cov.parent()
                + [
                    "children",
                    {
                        "id": "base_coverage",
                        "clamp": 1 if cov_range[0] == 0 else False,
                        "range": cov_range,
                    },
                    "children",
                ],
            )
        )
        parsed.append(
            Variable(
                "%s_read_cov" % cov_file_name,
                values=read_covs,
                meta={
                    "field_id": "%s_read_cov" % cov_file_name,
                    "file": cov_meta["f"],
                    "reads_mapped": cov_meta["reads_mapped"],
                    "reads_unmapped": cov_meta["reads_unmapped"],
                },
                parents=cov.parent()
                + [
                    "children",
                    {
                        "id": "read_coverage",
                        "datatype": "integer",
                        "clamp": 1 if read_cov_range[0] == 0 else False,
                        "range": read_cov_range,
                    },
                    "children",
                ],
            )
        )
    ranks = blob_db["dict_of_blobs"][identifiers.values[0]]["taxonomy"][
        blob_db["taxrules"][0]
    ].keys()
    for tax_rule in blob_db["taxrules"]:
        if "cat" not in kwargs["meta"].plot:
            kwargs["meta"].plot.update({"cat": "%s_phylum" % tax_rule})
        hit_list = hits_from_blob_db(blob_db, tax_rule)
        parsed.append(
            MultiArray(
                "%s_hits" % tax_rule,
                values=hit_list,
                meta={
                    "field_id": "%s_hits" % tax_rule,
                    "type": "multiarray",
                    "datatype": "mixed",
                    "preload": False,
                    "active": False,
                    "files": [m["f"] for x, m in blob_db["hitLibs"].items()],
                },
                parents=hits.parent() + ["children", {"id": tax_rule}, "children"],
                category_slot=None,
                headers=["taxid", "score"],
            )
        )
        for rank in ranks:
            field_id = "%s_%s" % (tax_rule, rank)
            parsed.append(
                Category(
                    field_id,
                    values=values[field_id],
                    meta={"field_id": field_id},
                    parents=hits.parent() + ["children", {"id": tax_rule}, "children"],
                )
            )
            parents = hits.parent() + [
                "children",
                {"id": tax_rule},
                "children",
                {"id": field_id},
                "data",
            ]
            field_id = "%s_%s_cindex" % (tax_rule, rank)
            parsed.append(
                Variable(
                    field_id,
                    values=values[field_id],
                    meta={
                        "scale": "scaleLinear",
                        "field_id": field_id,
                        "datatype": "integer",
                        "range": [min(values[field_id]), max(values[field_id])],
                        "preload": False,
                        "active": False,
                    },
                    parents=parents,
                )
            )
            field_id = "%s_%s_score" % (tax_rule, rank)
            _min = min(values[field_id])
            parsed.append(
                Variable(
                    field_id,
                    values=values[field_id],
                    meta={
                        "scale": "scaleLog",
                        "field_id": field_id,
                        "clamp": 1 if _min == 0 else False,
                        "datatype": "float",
                        "range": [_min, max(values[field_id])],
                        "preload": False,
                        "active": False,
                    },
                    parents=parents,
                )
            )

    return parsed


def parent():
    """Set standard metadata for synonyms."""
    return []



#!/usr/bin/env python3

# pylint: disable=no-member, too-many-branches, too-many-nested-blocks

"""
Remove fields from a BlobDir.

Usage:
    blobtools remove [--all] [--busco] [--cov] [--fasta] [--field STRING...] [--hits]
                     DIRECTORY

Arguments:
    DIRECTORY             Existing Blob directory.

Options:
    --all             Remove all fields except identifiers.
    --busco           Remove all BUSCO fields.
    --cov             Remove all cov and read_cov fields.
    --fasta           Remove gc, length and ncount fields.
    --field STRING    Remove fields by ID.
    --hits            Remove all taxonomy fields.

Examples:
    # 1. Remove BUSCO and ncount fields from a BlobDir
    ./blobtools remove --busco --field ncount BlobDir

"""

import glob
import sys
from collections import defaultdict

from docopt import docopt

from ..lib import busco
from ..lib import cov
from ..lib import fasta
from ..lib import file_io
from ..lib import hits
from .fetch import fetch_metadata
from .version import __version__

FIELDS = [
    {"flag": "--fasta", "module": fasta},
    {"flag": "--busco", "module": busco},
    {"flag": "--cov", "module": cov},
    {"flag": "--hits", "module": hits},
]


def remove_field(meta, field_id):
    """Delete field and any descendants."""
    field_ids = []
    if meta.has_field(field_id):
        field_ids = meta.remove_field(field_id)
    meta.plot = {
        key: value for key, value in meta.plot.items() if value not in field_ids
    }
    return field_ids


def remove_read_metadata(meta, field_ids):
    """Delete any read metadata for remove fields."""
    covs = defaultdict(dict)
    for field_id in field_ids:
        if field_id.endswith("_cov"):
            if field_id.endswith("_read_cov"):
                root = field_id.replace("_read_cov", "")
                covs[root].update({"read": True})
            else:
                root = field_id.replace("_cov", "")
                covs[root].update({"base": True})
    for key, value in covs.items():
        if "base" in value and "read" in value:
            meta.reads.pop(key, None)


def remove_static_plots(meta, directory):
    """Delete static plots."""
    files = glob.glob("%s/*.??g" % directory)
    for file in files:
        file_io.delete_file(file)
    setattr(meta, "static_plots", False)


def main(args):
    """Entrypoint for blobtools remove."""
    meta = fetch_metadata(args["DIRECTORY"], **args)
    field_ids = []
    for field in FIELDS:
        if args[field["flag"]] or args["--all"]:
            field_ids += field["module"].remove_from_meta(meta=meta)
    for field_id in args["--field"]:
        field_ids += remove_field(meta, field_id)
    if args["--all"]:
        for field_id in meta.list_fields():
            if field_id != "identifiers":
                field_ids += remove_field(meta, field_id)
    if meta.reads:
        remove_read_metadata(meta, field_ids)
    if meta.plot:
        axes = ["x", "y", "z", "cat"]
        for axis in axes:
            if axis not in meta.plot:
                remove_static_plots(meta, args["DIRECTORY"])
                break
    else:
        remove_static_plots(meta, args["DIRECTORY"])
    if field_ids:
        file_io.delete_file("%s/CHECKSUM" % args["DIRECTORY"])
        file_io.delete_file("%s/summary.json" % args["DIRECTORY"])
        for field_id in field_ids:
            file_io.delete_file("%s/%s.json" % (args["DIRECTORY"], field_id))
        file_io.write_file("%s/meta.json" % args["DIRECTORY"], meta.to_dict())


def cli():
    """Entry point."""
    if len(sys.argv) == sys.argv.index(__name__.split(".")[-1]) + 1:
        args = docopt(__doc__, argv=[])
    else:
        args = docopt(__doc__, version=__version__)
    main(args)


if __name__ == "__main__":
    cli()



#!/usr/bin/env python3
"""Dataset Class module."""

# pylint: disable=no-member, too-many-nested-blocks, too-many-branches, too-many-instance-attributes


class Metadata():
    """Class for field and dataset metadata."""

    __slots__ = ['dataset_id',
                 'assembly',
                 '_field_ids',
                 '_field_list',
                 'fields',
                 'links',
                 'name',
                 'origin',
                 'parents',
                 'plot',
                 'busco',
                 'reads',
                 'record_type',
                 'records',
                 'settings',
                 'similarity',
                 'static_plots',
                 'taxon',
                 'version',
                 'revision'
                 ]

    def __init__(self, dataset_id, **kwargs):
        """Init Dataset class."""
        self.dataset_id = dataset_id
        self.fields = []
        self.name = dataset_id
        self.links = {}
        self.assembly = {}
        self.plot = {}
        self.taxon = {}
        self.record_type = 'record'
        self.records = 0
        self.version = 1
        self.revision = 0
        self.update_data(**kwargs)
        self._field_list = self.list_fields()
        self._field_ids = list(self._field_list.keys())

    def update_data(self, **kwargs):
        """Update values and keys for an existing field."""
        for key, value in kwargs.items():
            if key == 'id':
                key = 'dataset_id'
            if key == 'assembly' and 'level' in value:
                setattr(self, 'record_type', value['level'])
            setattr(self, key, value)

    def list_fields(self):
        """List all fields in dataset."""
        field_ids = self._list_fields(self.fields)
        return field_ids

    def has_field(self, field_id):
        """Return true if a field_id is present."""
        return field_id in self._field_ids

    def remove_field(self, field_id):
        """Remove field and all descendants from dataset."""
        field_ids = []
        if self.has_field(field_id):
            meta = self.field_meta(field_id)
            field_ids = [field_id]
            field_ids += self.field_descendant_list(field_id, True)
            self.fields[:] = [d for d in self.fields if d.get('id') not in field_ids]
            self._field_list = self.list_fields()
            self._field_ids = list(self._field_list.keys())
            if meta.get('parent'):
                parent_meta = self.field_meta(meta['parent'])
                if 'children' in parent_meta:
                    parent_meta['children'][:] = [d for d in parent_meta['children'] if d.get('id') != field_id]
                if 'data' in parent_meta:
                    parent_meta['data'][:] = [d for d in parent_meta['data'] if d.get('id') != field_id]
                if not parent_meta.get('children', None) and not parent_meta.get('data', None):
                    field_ids += self.remove_field(parent_meta['id'])
        return field_ids

    def field_meta(self, field_id):
        """Return full metadata for a field."""
        if not self.has_field(field_id):
            return False
        meta = {key: value for key, value in self._field_list[field_id].items()}
        if meta.get('parent'):
            meta = self.add_parent_meta(meta, meta['parent'])
        return meta

    def field_descendant_list(self, field_id, full=False):
        """Return list of descendant fields for a field."""
        descendants = []
        meta = self.field_meta(field_id)
        if meta.get('children'):
            for child in meta['children']:
                descendants += self.field_descendant_list(child['id'])
                if full:
                    descendants.append(child['id'])
        else:
            descendants.append(meta['id'])
            if meta.get('data'):
                for child in meta['data']:
                    descendants.append(child['id'])
        return descendants

    def field_parent_list(self, field_id):
        """Return list of parent fields for a field."""
        parents = []
        meta = self.field_meta(field_id)
        if meta.get('parent'):
            parent_meta = self.field_meta(meta['parent'])
            parents.append({'id': meta['parent']})
            for key, value in parent_meta.items():
                if key not in ('field_id', 'children', 'data'):
                    parents[-1].update({key: value})
            if 'children' in parent_meta:
                parents.append('children')
            elif 'data' in parent_meta:
                parents.append('data')
            parents = self.field_parent_list(meta['parent']) + parents
        return parents

    def add_parents(self, parents):
        """Add field metadata."""
        fields = self.fields
        if not parents:
            return fields
        parent = self._field_list
        for required in parents:
            if isinstance(required, dict):
                if isinstance(parent, dict):
                    if required['id'] not in parent:
                        parent.update({required['id']: required})
                        fields.append(required)
                    else:
                        for key in required.keys():
                            if key not in ['children', 'data']:
                                parent[required['id']][key] = required[key]
                    parent = parent[required['id']]
                elif required['id'] in self._field_list:
                    for key in required.keys():
                        if key not in ['children', 'data']:
                            self._field_list[required['id']][key] = required[key]
                    parent = self._field_list[required['id']]
                elif isinstance(parent, list):
                    try:
                        index = next(i for i, field in enumerate(parent)
                                     if field['id'] == required['id'])
                        parent = parent[index]
                        for key in required.keys():
                            if key not in ['children', 'data']:
                                parent[key] = required[key]
                    except StopIteration:
                        parent.append(required)
                        parent = required
            else:
                if required not in parent:
                    parent.update({required: []})
                parent = parent[required]
        return parent

    def add_field(self, parents=None, **kwargs):
        """Add field metadata."""
        if parents is None:
            parents = []
        parent = self.add_parents(parents)
        if self.has_field(kwargs['field_id']):
            meta = self._field_list[kwargs['field_id']]
        else:
            index = next((i for (i, d) in enumerate(parent) if d['id'] == kwargs['field_id']), None)
            if index is not None:
                meta = parent[index]
            else:
                meta = {}
                parent.append(meta)
        for key, value in kwargs.items():
            if key == 'field_id':
                key = 'id'
            meta[key] = value
        # if parents:
        #     meta['parents'] = parents

    def to_dict(self):
        """Create a dict of metadata."""
        data = {}
        for key in self.__slots__:
            if not key.startswith('_'):
                if hasattr(self, key):
                    if key == 'dataset_id':
                        data['id'] = getattr(self, key)
                    else:
                        data[key] = getattr(self, key)
        return data

    @staticmethod
    def _list_fields(parent, fields=None, parent_id=None):
        """Create a dict of fields."""
        if fields is None:
            fields = {}
        for field in parent:
            if parent_id:
                field.update({'parent': parent_id})
            fields.update({field['id']: field})
            if 'children' in field:
                Metadata._list_fields(field['children'], fields, field['id'])
            if 'data' in field:
                Metadata._list_fields(field['data'], fields, field['id'])
        return fields

    def add_parent_meta(self, meta, parent_id):
        """Add parent metadata to field metadata."""
        parent_meta = self.field_meta(parent_id)
        for key, value in parent_meta.items():
            if key not in list(meta.keys()) + ['children', 'data']:
                meta.update({key: value})
        if parent_meta.get('parent'):
            meta = self.add_parent_meta(meta, parent_meta['parent'])
        return meta


if __name__ == '__main__':
    import doctest
    doctest.testmod()



#!/usr/bin/env python3
# pylint: skip-file

"""Parse FASTA sequence into Fields."""

import pathlib
import statistics
from collections import Counter
from collections import OrderedDict

from tqdm import tqdm

from ..lib import file_io
from .field import Identifier
from .field import Variable
from .run_external import seqtk_subseq


def base_composition(seq_str):
    """Sequence base composition summary."""
    counts = Counter(seq_str.upper())
    at_bases = ["A", "T", "W"]
    gc_bases = ["C", "G", "S"]
    at_count = 0
    gc_count = 0
    for base in at_bases:
        at_count += counts[base]
    for base in gc_bases:
        gc_count += counts[base]
    acgt_count = at_count + gc_count
    gc_portion = float("%.4f" % (gc_count / acgt_count))
    n_count = len(seq_str) - acgt_count
    return gc_portion, n_count


def apply_filter(ids, fasta_file, **kwargs):
    """Filter FASTA format assembly."""
    suffix = kwargs["--suffix"]
    path = pathlib.Path(fasta_file)
    outfile = path.parent / (path.stem + "." + suffix + path.suffix)
    seqtk_subseq(fasta_file, "\n".join(ids), outfile)


def parse(file, **kwargs):
    """Parse all synonym files."""
    parsed = []
    _lengths = OrderedDict()
    _gc_portions = OrderedDict()
    _n_counts = OrderedDict()
    lengths = []
    gc_portions = []
    n_counts = []
    print("Loading sequences from %s" % file)
    pbar = tqdm(file_io.stream_fasta(file))
    for seq_id, seq_str in pbar:
        pbar.set_description(" - processing %s" % seq_id)
        _lengths[seq_id] = len(seq_str)
        _gc_portions[seq_id], _n_counts[seq_id] = base_composition(seq_str)
    identifiers = kwargs["dependencies"]["identifiers"]
    if not identifiers:
        identifiers = Identifier(
            "identifiers",
            meta={"field_id": "identifiers"},
            values=list(_lengths.keys()),
            parents=[],
        )
        kwargs["meta"].assembly.update({"scaffold-count": len(identifiers.values)})
        parsed.append(identifiers)
    for seq_id in identifiers.values:
        lengths.append(_lengths[seq_id] if seq_id in _lengths else 0)
        gc_portions.append(_gc_portions[seq_id] if seq_id in _gc_portions else 0)
        n_counts.append(_n_counts[seq_id] if seq_id in _n_counts else 0)
    kwargs["meta"].assembly.update({"span": sum(lengths)})
    parsed.append(
        Variable(
            "gc",
            meta={
                "field_id": "gc",
                "preload": True,
                "scale": "scaleLinear",
                "name": "GC",
                "datatype": "float",
                "range": [min(gc_portions), max(gc_portions)],
            },
            values=gc_portions,
            parents=[],
        )
    )
    _min = min(lengths)
    parsed.append(
        Variable(
            "length",
            meta={
                "preload": True,
                "scale": "scaleLog",
                "field_id": "length",
                "name": "Length",
                "clamp": 1 if _min == 0 else False,
                "datatype": "integer",
                "range": [_min, max(lengths)],
            },
            values=lengths,
            parents=[],
        )
    )
    parsed.append(
        Variable(
            "ncount",
            meta={
                "scale": "scaleLinear",
                "field_id": "ncount",
                "name": "N count",
                "datatype": "integer",
                "range": [min(n_counts), max(n_counts)],
            },
            values=n_counts,
            parents=[],
        )
    )
    if "x" not in kwargs["meta"].plot:
        kwargs["meta"].plot.update({"x": "gc"})
    if "z" not in kwargs["meta"].plot:
        kwargs["meta"].plot.update({"z": "length"})
    return parsed


def parent():
    """Set standard metadata for synonyms."""
    return []


def summarise(indices, fields, **kwargs):
    """Summarise assembly sequence stats."""
    gcs = [fields["gc"].values[i] for i in indices]
    lengths = [fields["length"].values[i] for i in indices]
    gc_mean = sum([gc * length for gc, length in zip(gcs, lengths)]) / sum(lengths)
    gc = float("%.4f" % gc_mean)
    at = 1 - gc
    if "ncount" in fields and fields["ncount"]:
        ncount = sum([fields["ncount"].values[i] for i in indices])
    else:
        ncount = 0
    span = sum(lengths)
    return {"at": at, "gc": gc, "n": float("%.4f" % (ncount / span))}


def remove_from_meta(meta):
    """Delete all fasta-derived fields."""
    field_ids = []
    if meta.has_field("gc"):
        field_ids += meta.remove_field("gc")
        meta.plot.pop("x", None)
    if meta.has_field("length"):
        field_ids += meta.remove_field("length")
        meta.plot.pop("z", None)
    if meta.has_field("ncount"):
        field_ids += meta.remove_field("ncount")
    meta.assembly.pop("file", None)
    return field_ids



#!/usr/bin/env python3
# pylint: disable=too-many-locals,too-many-branches,too-many-arguments,too-many-statements,too-many-nested-blocks

"""Parse text files into fields."""

import math
import pathlib
import re
from collections import defaultdict

from ..lib import file_io
from .field import Array
from .field import Category
from .field import MultiArray
from .field import Variable


def parse_header_row(delimiter, header_row, columns):
    """Parse header row."""
    head = re.split(delimiter, header_row)
    for i, col in enumerate(columns):
        try:
            index, name = col.split("=")
            if not index.isdigit():
                try:
                    index = head.index(index)
                    columns[i] = "%d=%s" % (index + 1, name)
                except ValueError:
                    pass
        except ValueError:
            if col.isdigit():
                try:
                    columns[i] = "%s=%s" % (col, head[int(col) - 1])
                except IndexError:
                    exit("ERROR: column index out of range '%s'" % col)
            else:
                try:
                    index = head.index(col)
                    columns[i] = "%s=%s" % (str(index + 1), col)
                except ValueError:
                    pass
    if not columns:
        return head
    return columns


def map_fields(delimiter, first_row, columns):
    """Set column headers and field types."""
    sample_data = re.split(delimiter, first_row)
    width = len(sample_data)
    cols = []
    headers = {}
    types = {}
    for i, col in enumerate(columns):
        if col:
            try:
                index, name = col.split("=")
                if index.isdigit() and width >= int(index):
                    cols.append(int(index) - 1)
                    headers[int(index) - 1] = name
                else:
                    exit("ERROR: column index out of range '%s'" % col)
            except ValueError:
                if col.isdigit():
                    exit("ERROR: no name specified for column %s" % col)
                else:
                    cols.append(i)
                    headers[i] = col
    for col in cols:
        if re.search("identifier", headers[col], re.IGNORECASE):
            types[headers[col]] = "Identifier"
        else:
            datum = sample_data[col]
            try:
                float(datum)
                types[headers[col]] = "Variable"
            except ValueError:
                types[headers[col]] = "Category"
    return cols, headers, types, width


def parse_rows(delimiter, lines, width, no_array, cols, types, headers):
    """Parse rows and test for duplicate identifiers."""
    ids = set()
    rows = []
    id_rows = defaultdict(list)
    index = 0
    array = False
    for line in lines:
        line = line.replace('"', "")
        data = re.split(delimiter, line)
        if len(data) < width:
            continue
        row = {}
        for col in cols:
            if types[headers[col]] == "Identifier":
                if data[col] in ids:
                    if no_array:
                        exit(
                            "ERROR: found multiple instances of Identifier '%s'"
                            % data[col]
                        )
                    array = True
                ids.add(data[col])
                id_rows[data[col]].append(index)
                index += 1
            else:
                row[headers[col]] = data[col]
        rows.append(row)
    return rows, id_rows, array, ids


def rows_to_results(rows, id_rows, types, array, field_name):
    """Make fields from rows."""
    field_names = [type for type in types.keys() if types[type] != "Identifier"]
    if field_name:
        results = {field_name: {}}
    else:
        results = {name: {} for name in field_names}
    for ident, indices in id_rows.items():
        if field_name:
            if len(field_names) > 1:
                results[field_name][ident] = []
        elif array:
            for name in field_names:
                results[name][ident] = []
        for index in indices:
            row = rows[index]
            if field_name:
                data = [row[name] for name in field_names]
                if len(field_names) == 1:
                    data = data[0]
                if array:
                    results[field_name][ident].append(data)
                else:
                    results[field_name][ident] = data
            else:
                for name in field_names:
                    if array:
                        results[name][ident].append([row[name]])
                    else:
                        results[name][ident] = row[name]
    return results


def results_to_fields(results, types, cols, headers, text_file, delimiter, identifiers):
    """Convert results to fields."""
    fields = []
    field_types = {
        "Variable": Variable,
        "Category": Category,
        "Array": Array,
        "MultiArray": MultiArray,
    }
    array_type = "array"
    for col in cols:
        if types[headers[col]] == "Identifier":
            id_column = col
        else:
            if array_type and types[headers[col]] != array_type:
                array_type = "mixed"
                break
            array_type = "string" if types[headers[col]] == "Category" else "float"
    for key, values in results.items():
        ident, sample = next(iter(values.items()))
        blank = "NA" if types[headers[col]] == "Category" else 0
        if not identifiers.validate_list(list(results[key].keys())):
            print(
                "WARN: Contig names in the text file did not match dataset identifiers."
            )
        kwargs = {
            "meta": {
                "field_id": key,
                "name": key,
                "preload": False,
                "active": False,
                "file": text_file,
                "id_column": id_column,
                "delimiter": delimiter,
                "datatype": array_type,
            },
            "parents": [],
        }
        if isinstance(sample, list):
            blank = []
            array_headers = []
            index = -1
            if key in types:
                array_headers.append(key)
                if types[key] == "Category" and "category_slot" not in kwargs:
                    kwargs.update({"category_slot": 0})
            else:
                for col in cols:
                    if types[headers[col]] != "Identifier":
                        index += 1
                        array_headers.append(headers[col])
                        if (
                            types[headers[col]] == "Category"
                            and "category_slot" not in kwargs
                        ):
                            kwargs.update({"category_slot": index})
            kwargs.update({"headers": array_headers})
            if isinstance(sample[0], list):
                field_type = field_types["MultiArray"]
                kwargs.update({"type": "multiarray"})
            else:
                field_type = field_types["Array"]
                kwargs.update({"type": "array"})
        else:
            field_type = field_types[types[key]]
            kwargs.update({"type": types[key].lower()})

        if kwargs["type"] == "variable":
            vals = []
            is_float = False
            for ident in identifiers.values:
                value = results[key][ident] if ident in results[key] else blank
                vals.append(value)
                if kwargs["type"] == "variable" and not is_float:
                    try:
                        int(value)
                    except ValueError:
                        is_float = True
            min_max = [math.inf, -math.inf]
            values = []
            for value in vals:
                value = float(value) if is_float else int(value)
                values.append(value)
                min_max = [min(min_max[0], value), max(min_max[1], value)]
            kwargs["meta"].update({"range": min_max})
            if is_float:
                kwargs["meta"].update({"datatype": "float"})
            else:
                kwargs["meta"].update({"datatype": "integer"})
            if min_max[0] < 0 or min_max[0] > min_max[1] / 1000:
                kwargs["meta"].update({"scale": "scaleLinear"})
            else:
                kwargs["meta"].update({"scale": "scaleLog"})
                if min_max[0] == 0:
                    kwargs["meta"].update({"clamp": 0.01})
        else:
            values = [
                results[key][ident] if ident in results[key] else blank
                for ident in identifiers.values
            ]
            if kwargs["type"] == "category":
                kwargs["meta"].update({"datatype": "string"})
        field = field_type(key, values=values, **kwargs)
        fields.append(field)
    return fields


def set_delimiter(delimiter, *, sample=None):
    """Set text delimiter."""
    if delimiter == "whitespace":
        if sample is not None:
            if "\t" in sample:
                return re.compile(r"\t")
        return re.compile(r"\s+")
    else:
        return re.compile(r"%s" % delimiter)


def parse_text(text_file, delimiter, columns, header, no_array, identifiers):
    """Parse text file into Category and/or Variable fields."""
    try:
        text_file, field_name = text_file.split("=")
    except ValueError:
        field_name = False
    data = file_io.read_file(text_file)
    lines = data.split("\n")
    delimit = set_delimiter(delimiter, sample=lines[0])
    if columns:
        columns = columns.split(",")
    else:
        columns = []
    if header:
        header_row = lines.pop(0).replace('"', "")
        columns = parse_header_row(delimit, header_row, columns)
    cols, headers, types, width = map_fields(
        delimit, lines[0].replace('"', ""), columns
    )
    rows, id_rows, array = parse_rows(
        delimit, lines, width, no_array, cols, types, headers
    )[:3]
    # if not identifiers.validate_list(list(ids)):
    #     exit('ERROR: contig names in the text file did not match dataset identifiers.')
    results = rows_to_results(rows, id_rows, types, array, field_name)
    fields = results_to_fields(
        results, types, cols, headers, text_file, delimiter, identifiers
    )
    # meta = {'file': text_file}
    return fields
    # results = defaultdict(list)
    # for line in lines:
    #     if header:
    #         row = re.split(' +', line)
    #         if len(row) > 1:
    #             if row[1].startswith('v.'):
    #                 meta.update({'version': row[1]})
    #             elif row[1] == 'Mode:':
    #                 meta.update({'mode': row[2]})
    #                 meta.update({'field_id': "trnascan_%s" % row[2].lower()})
    #             elif row[1].startswith('------'):
    #                 header = False
    #     else:
    #         row = re.split(r' +|\t', line)
    #         if len(row) == 9:
    #             results[row[0]].append([row[4], row[5]])
    # if not identifiers.validate_list(list(results.keys())):
    #     raise UserWarning('Contig names in the tRNAScan file did not match dataset identifiers.')
    # values = [results[id] if id in results else [] for id in identifiers.values]
    # trnascan_field = MultiArray(meta['field_id'],
    #                             values=values,
    #                             meta=meta,
    #                             headers=('tRNA_type', 'Anticodon'),
    #                             parents=['children']
    #                             )
    # return trnascan_field


def apply_filter(ids, text_file, **kwargs):
    """Filter Text file."""
    suffix = kwargs["--suffix"]
    path = pathlib.Path(text_file)
    outfile = str(path.parent / (path.stem + "." + suffix + path.suffix))
    data = file_io.read_file(text_file)
    lines = data.split("\n")
    delimiter = kwargs["--text-delimiter"]
    delimit = set_delimiter(delimiter, sample=lines[0])
    id_col = int(kwargs["--text-id-column"]) - 1
    output = []
    if kwargs["--text-header"]:
        header_row = lines.pop(0)
        header_row.rstrip()
        output.append(header_row)
    for line in lines:
        line = line
        row = re.split(delimit, line.replace('"', ""))
        try:
            if row[id_col] in ids:
                output.append(line)
        except IndexError:
            output.append(line)
    file_io.write_file(outfile, output, plain=True)


def parse(files, **kwargs):
    """Parse all text files."""
    parsed = []
    for file in files:
        try:
            fileData = parse_text(
                file,
                delimiter=kwargs["--text-delimiter"],
                columns=kwargs["--text-cols"],
                header=kwargs["--text-header"],
                no_array=kwargs["--text-no-array"],
                identifiers=kwargs["dependencies"]["identifiers"],
            )
        except AttributeError:
            continue
        parsed = parsed + fileData
    return parsed


def parent():
    """Set standard metadata for text."""
    return []



#!/usr/bin/env python3
# pylint: skip-file

"""Parse BED files into Fields."""


import math
import sys
from collections import Counter
from collections import OrderedDict
from collections import defaultdict
from copy import deepcopy
from glob import glob
from itertools import zip_longest
from os import path
from os import stat
from pathlib import Path

from tolkein import tofile
from tqdm import tqdm

from ..lib import file_io
from .field import Array
from .field import Identifier
from .field import MultiArray
from .field import Variable
from .run_external import seqtk_subseq


def field_settings():
    """Field settings for bed file imports."""
    return {
        "gc": {
            "meta": {
                "preload": True,
                "name": "GC",
                "scale": "scaleLinear",
                "type": "variable",
                "datatype": "float",
            },
            "parents": [
                {
                    "id": "gc_data",
                    "scale": "scaleLinear",
                    "type": "variable",
                    "datatype": "float",
                },
                "children",
            ],
            "plot_axis": "x",
        },
        "length": {
            "meta": {
                "preload": True,
                "name": "Length",
                "clamp": 1,
                "scale": "scaleLog",
                "datatype": "integer",
                "type": "variable",
            },
            "parents": [
                {
                    "id": "length_data",
                    "scale": "scaleLog",
                    "datatype": "integer",
                    "type": "variable",
                },
                "children",
            ],
            "plot_axis": "z",
        },
        "position": {
            "meta": {
                "name": "Position",
                "scale": "scaleLinear",
                "datatype": "integer",
                "type": "variable",
                "range": [0, 1],
            },
            "parents": [
                {
                    "id": "position_data",
                    "scale": "scaleLinear",
                    "datatype": "integer",
                    "type": "variable",
                },
                "children",
            ],
        },
        "proportion": {
            "meta": {
                "name": "Proportion",
                "scale": "scaleLinear",
                "datatype": "float",
                "type": "variable",
                "range": [0, 1],
            },
            "parents": [
                {
                    "id": "proportion_data",
                    "scale": "scaleLinear",
                    "datatype": "float",
                    "type": "variable",
                },
                "children",
            ],
        },
        "cov": {
            "meta": {
                "preload": 1,
                "scale": "scaleLog",
                "name": "Coverage",
                "clamp": 0.01,
                "datatype": "float",
                "type": "variable",
                "range": [math.inf, -math.inf],
            },
            "parents": [
                {
                    "id": "coverage",
                    "type": "variable",
                    "datatype": "float",
                },
                "children",
            ],
            "plot_axis": "y",
        },
        "n": {
            "meta": {
                "name": "N",
                "scale": "scaleLinear",
                "datatype": "float",
                "type": "variable",
            },
            "parents": [
                {
                    "id": "n_data",
                    "scale": "scaleLinear",
                    "datatype": "float",
                    "type": "variable",
                },
                "children",
            ],
        },
        "masked": {
            "meta": {
                "name": "Masked",
                "scale": "scaleLinear",
                "datatype": "float",
                "type": "variable",
            },
            "parents": [
                {
                    "id": "masked_data",
                    "scale": "scaleLinear",
                    "datatype": "float",
                    "type": "variable",
                },
                "children",
            ],
        },
        "ncount": {
            "meta": {
                "name": "N count",
                "scale": "scaleLinear",
                "datatype": "integer",
                "type": "variable",
            },
            "parents": [
                {
                    "id": "ncount_data",
                    "scale": "scaleLinear",
                    "datatype": "integer",
                    "type": "variable",
                },
                "children",
            ],
        },
        "count": {
            "meta": {
                "scale": "scaleLinear",
                "name": "Count",
                "datatype": "integer",
                "type": "variable",
            },
            "parents": [
                {
                    "id": "count_data",
                    "type": "variable",
                    "datatype": "integer",
                },
                "children",
            ],
        },
        "cpm": {
            "meta": {
                "scale": "scaleLinear",
                "name": "Count per Mb",
                "datatype": "float",
                "type": "variable",
            },
            "parents": [
                {
                    "id": "cpm_data",
                    "type": "variable",
                    "datatype": "float",
                },
                "children",
            ],
        },
    }


def parse_full_bed(filename):
    """Parse a BED file containing one value per sequence."""
    parsed = {}
    with tofile.open_file_handle(filename) as fh:
        for line in fh:
            row = line.strip().split("\t")
            if len(row) < 5:
                return parsed
            parsed[row[0]] = float(row[4])
    return parsed


def parse_window_bed(filename):
    """Parse a BED file containing multiple values per sequence."""
    parsed = defaultdict(list)
    with tofile.open_file_handle(filename) as fh:
        for line in fh:
            row = line.strip().split("\t")
            parsed[row[0]].append((int(row[1]), float(row[4])))
    windowed = {}
    for seq_id, arr in parsed.items():
        windowed[seq_id] = [tup[1] for tup in sorted(arr, key=lambda tup: tup[0])]
    return windowed


def parse_full_tsv(filename):
    """Parse a TSV file containing one value per sequence."""
    values = defaultdict(dict)
    sd = defaultdict(dict)
    n = defaultdict(dict)
    header = None
    with tofile.open_file_handle(filename) as fh:
        for line in fh:
            row = line.strip().split("\t")
            if header is None:
                header = {key: idx + 3 for idx, key in enumerate(row[3:])}
                continue
            length = int(row[2]) - int(row[1])
            values["length"][row[0]] = length
            values["position"][row[0]] = int(row[2])
            for key, idx in header.items():
                if key.endswith("_sd"):
                    sd[key[: key.rfind("_sd")]][row[0]] = float(row[idx])
                elif key.endswith("_n"):
                    n[key[: key.rfind("_n")]][row[0]] = float(row[idx])
                elif key.endswith("_cpm"):
                    values[key][row[0]] = float(
                        "%.3g" % (float(row[idx]) / length * 1000000)
                    )
                elif key.endswith("_count"):
                    values[key][row[0]] = int(row[idx])
                else:
                    values[key][row[0]] = float(row[idx])
    return values, sd, n


def parse_windowed_tsv(filename, window):
    """Parse a TSV file containing one value per sequence."""
    values = defaultdict(lambda: defaultdict(list))
    sd = defaultdict(lambda: defaultdict(list))
    n = defaultdict(lambda: defaultdict(list))
    lengths = {}
    header = None
    with tofile.open_file_handle(filename) as fh:
        for line in fh:
            row = line.strip().split("\t")
            if header is None:
                header = {key: idx + 3 for idx, key in enumerate(row[3:])}
                continue
            length = int(row[2]) - int(row[1])
            if float(window) > 1:
                prev_len = lengths.get(row[0], 0)
                if length < prev_len:
                    continue
                lengths[row[0]] = length
            values["length"][row[0]].append(length)
            values["position"][row[0]].append(round(int(row[1]) + length / 2))
            for key, idx in header.items():
                if key.endswith("_sd"):
                    sd[key[: key.rfind("_sd")]][row[0]].append(float(row[idx]))
                elif key.endswith("_n"):
                    n[key[: key.rfind("_n")]][row[0]].append(float(row[idx]))
                elif key.endswith("_cpm"):
                    values[key][row[0]].append(
                        float("%.3g" % (float(row[idx]) / length * 1000000))
                    )
                elif key.endswith("_count"):
                    values[key][row[0]].append(int(row[idx]))
                else:
                    values[key][row[0]].append(float(row[idx]))
    return values, sd, n


def parse_tsvfiles(files):
    """Parse all tsvfiles."""
    windows = {}
    windows_sd = {}
    windows_n = {}
    [fullfile, *files] = sorted(files, key=len)
    full, sd, n = parse_full_tsv(fullfile)
    for file in files:
        if stat(file).st_size == 0:
            continue
        window = "".join(y for x, y in zip_longest(fullfile, file) if x != y).replace(
            ".tsv", ""
        )
        windows[window], windows_sd[window], windows_n[window] = parse_windowed_tsv(
            file, window
        )
    return (
        fullfile,
        {"values": windows, "sd": windows_sd, "n": windows_n},
        {"values": full, "sd": sd, "n": n},
    )


def parse_bedfiles(files):
    """Parse all bedfiles."""
    full = {}
    windows = {"0.1": {}}
    filenames = {}
    for filename in files:
        filepath = Path(filename)
        field = filepath.name.split(".")[-2]
        filenames[field] = filepath.name
        if field.endswith("_windows"):
            windows["0.1"][field[: field.rfind("_windows")]] = parse_window_bed(
                filename
            )
        else:
            full[field] = parse_full_bed(filename)
    return filenames, windows, full


def validate_range(meta):
    """Ensure range is greater than zero."""
    if meta["range"][1] < meta["range"][0]:
        return False
    if meta["range"][1] == meta["range"][0]:
        if meta["datatype"] == "integer":
            meta["range"][1] += 1
        elif meta["range"][1] == 0:
            meta["range"][1] += 0.1
        else:
            mag = math.floor(math.log10(meta["range"][1]))
            meta["range"][1] += 10 ** (mag - 1)
    return True


def parse(files, **kwargs):
    if "--bedtsvdir" in kwargs or "--bedtsvdir" in kwargs:
        if isinstance(files, str) and path.isdir(files):
            print("Reading all TSV files in %s" % files)
            files = glob("%s/*.tsv" % files)
        filename, all_windows, full = parse_tsvfiles(files)
        filenames = {"all": filename}
    else:
        if isinstance(files, str) and path.isdir(files):
            print("Reading all BED files in %s" % files)
            files = glob("%s/*.bed" % files)
        filenames, all_windows, full = parse_bedfiles(files)
    full_n = {}
    full_sd = {}
    if isinstance(full, dict):
        full_sd = full["sd"]
        full_n = full["n"]
        full = full["values"]
    all_windows_n = {}
    all_windows_sd = {}
    if isinstance(all_windows, dict):
        all_windows_n = all_windows["n"]
        all_windows_sd = all_windows["sd"]
        all_windows = all_windows["values"]
    parsed = []
    settings = field_settings()
    identifiers = kwargs["dependencies"]["identifiers"]
    keys = []
    if "length" in full:
        keys = list(full["length"].keys())
        lengths = list(full["length"].values())
        kwargs["meta"].assembly.update({"span": sum(lengths)})
        if "z" not in kwargs["meta"].plot:
            kwargs["meta"].plot.update({"z": "length"})
    if "gc" in full and "x" not in kwargs["meta"].plot:
        kwargs["meta"].plot.update({"x": "gc"})
    if not identifiers:
        if not keys:
            print("ERROR: Unable to set identifiers")
            sys.exit(1)
        identifiers = Identifier(
            "identifiers",
            meta={"field_id": "identifiers"},
            values=keys,
            parents=[],
        )
        kwargs["meta"].assembly.update({"scaffold-count": len(identifiers.values)})
        parsed.append(identifiers)
    ranges = {
        key: {"range": [math.inf, -math.inf], "meta": {}} for key in settings.keys()
    }
    for field, data in full.items():
        filename = filenames.get(field, filenames.get("all", ""))
        if data:
            values = []
            for seq_id in identifiers.values:
                values.append(data[seq_id] if seq_id in data else 0)
            if values:
                meta = {}
                parents = []
                suffix = field.split("_")[-1]
                if suffix in settings:
                    meta = deepcopy(settings[suffix]["meta"])
                    meta.update(
                        {
                            "field_id": field,
                            "file": filename,
                        }
                    )
                    if meta["datatype"] == "integer":
                        values = [int(value) for value in values]
                    value_range = [min(values), max(values)]
                    if "clamp" in meta and value_range[0] >= meta["clamp"]:
                        meta["clamp"] = False
                    parent_range = False
                    if "parents" in settings[suffix]:
                        parents = settings[suffix]["parents"]
                        for parent in parents:
                            if "range" in parent:
                                parent_range = True
                                parent["range"][0] = min(
                                    parent["range"][0], value_range[0]
                                )
                                parent["range"][1] = max(
                                    parent["range"][1], value_range[1]
                                )
                    if not parent_range:
                        if "range" in meta:
                            meta["range"][0] = min(meta["range"][0], value_range[0])
                            meta["range"][1] = max(meta["range"][1], value_range[1])
                        else:
                            meta["range"] = value_range
                    if not validate_range(meta):
                        continue
                    if "preload" in meta and meta["preload"] == 1:
                        if value_range[1] > ranges[suffix]["range"][1]:
                            meta["preload"] = True
                            if "plot_axis" in settings[suffix]:
                                kwargs["meta"].plot.update(
                                    {settings[suffix]["plot_axis"]: field}
                                )
                            if "preload" in ranges[suffix]["meta"]:
                                ranges[suffix]["meta"]["preload"] = False
                            ranges[suffix].update({"range": value_range, "meta": meta})
                        else:
                            meta["preload"] = False
                    if field.endswith("_%s" % suffix):
                        meta["name"] = "%s %s" % (
                            field[: field.rfind("_%s" % suffix)],
                            meta["name"],
                        )
                parsed.append(
                    Variable(
                        field,
                        meta=meta,
                        values=values,
                        parents=parents,
                    )
                )
                if field in full_sd:
                    stats_values = []
                    for seq_id in identifiers.values:
                        values = (
                            [full_sd[field][seq_id], full_n[field][seq_id]]
                            if seq_id in data
                            else []
                        )
                        stats_values.append(values)
                    parsed.append(
                        Array(
                            "%s_stats" % field,
                            meta={
                                "field_id": "%s_stats" % field,
                                "name": "%s stats" % meta["name"],
                                "type": "array",
                                "datatype": "mixed",
                            },
                            values=stats_values,
                            parents=parents,
                            headers=["sd", "n"],
                        )
                    )
                for window, windows in all_windows.items():
                    windows_sd = all_windows_sd.get(window, {})
                    windows_n = all_windows_n.get(window, {})
                    if field in windows:
                        window_values = []
                        headers = [field]
                        if field in windows_sd:
                            headers += ["sd", "n"]
                        for seq_id in identifiers.values:
                            seq_values = []
                            if seq_id in data:
                                for idx, value in enumerate(windows[field][seq_id]):
                                    if meta["datatype"] == "integer":
                                        value = int(value)
                                    if field in windows_sd:
                                        value = [
                                            value,
                                            windows_sd[field][seq_id][idx],
                                            windows_n[field][seq_id][idx],
                                        ]
                                    else:
                                        value = [value]
                                    seq_values.append(value)
                            window_values.append(seq_values)
                        windows_field = "%s_windows" % field
                        if str(window) != "0.1":
                            windows_field += "_%s" % str(window)
                        parsed.append(
                            MultiArray(
                                windows_field,
                                meta={
                                    "field_id": windows_field,
                                    "name": "%s windows %s" % (meta["name"], window),
                                    "type": "multiarray",
                                    "datatype": "mixed",
                                },
                                values=window_values,
                                parents=parents,
                                headers=headers,
                            )
                        )
    return parsed


def parent():
    """Set standard metadata."""
    return []



#!/usr/bin/env python3

# pylint: disable=no-member, too-many-locals
# pylint: skip-file

"""Parse base and read coverage into Variable Field."""

import math
import os
import sys
from collections import defaultdict
from multiprocessing import Pool
from pathlib import Path

import pysam
from tqdm import tqdm

from .field import Variable
from .file_io import load_yaml
from .run_external import seqtk_subseq


def check_mapped_read(read):
    """Check read mapping flags."""
    if read.is_unmapped or read.is_secondary or read.is_supplementary:
        return False
    return True


def calculate_coverage(aln, reads_mapped):
    """Calculate base and read coverage."""
    _base_cov_dict = defaultdict(list)
    read_cov_dict = defaultdict(lambda: 0)
    allowed_operations = set([0, 7, 8])
    with tqdm(total=reads_mapped, unit_scale=True) as pbar:
        for read in aln.fetch(until_eof=True):
            if not check_mapped_read(read):
                continue
            for operation, length in read.cigartuples:
                if operation in allowed_operations:
                    _base_cov_dict[read.reference_name].append(length)
            read_cov_dict[read.reference_name] += 1
            pbar.update()
    base_cov_dict = {
        ref_name: sum(_base_cov) for ref_name, _base_cov in _base_cov_dict.items()
    }
    return base_cov_dict, read_cov_dict


def parse_bam(bam_file, **kwargs):
    """Parse coverage into a Variables."""
    identifiers = kwargs["dependencies"]["identifiers"]
    ids = identifiers.values
    lengths = kwargs["dependencies"]["length"].values
    ncounts = kwargs["dependencies"]["ncount"].values
    parts = bam_file.split("=")
    base_name = parts[1]
    f_char = Path(parts[0]).suffix[1]
    index_file = Path("%s.csi" % parts[0])
    if not index_file.is_file():
        pysam.index("-c", "-m", "14", parts[0])
    else:
        index_file = False
    stats = {}
    print("Loading mapping data from %s as %s" % (parts[0], parts[1]))
    with pysam.AlignmentFile(parts[0], "r%s" % f_char) as aln:
        stats = {"mapped": aln.mapped, "unmapped": aln.unmapped}
        _covs, _read_covs = calculate_coverage(aln, aln.mapped)
    if index_file:
        os.remove(index_file)
    if not identifiers.validate_list(list(_covs.keys())):
        raise UserWarning(
            "Contig names in the coverage file did not match dataset identifiers."
        )
    covs = []
    read_covs = []
    for index, seq_id in enumerate(ids):
        acgt_count = lengths[index] - ncounts[index]
        covs.append(
            float("%.4f" % (_covs[seq_id] / acgt_count)) if seq_id in _covs else 0
        )
        read_covs.append(_read_covs[seq_id] if seq_id in _read_covs else 0)
    field_id = "%s_cov" % base_name
    fields = {"cov_id": field_id}
    fields["cov_range"] = [
        min(covs + [kwargs["cov_range"][0]]),
        max(covs + [kwargs["cov_range"][1]]),
    ]
    fields["cov"] = Variable(
        field_id,
        values=covs,
        meta={"field_id": field_id, "file": bam_file},
        parents=[
            "children",
            {
                "id": "base_coverage",
                "clamp": 0.01 if fields["cov_range"][0] == 0 else False,
                "range": fields["cov_range"],
            },
            "children",
        ],
    )
    field_id = "%s_read_cov" % base_name
    fields["read_cov_range"] = [
        min(read_covs + [kwargs["read_cov_range"][0]]),
        max(read_covs + [kwargs["read_cov_range"][1]]),
    ]
    fields["read_cov"] = Variable(
        field_id,
        values=read_covs,
        meta={
            "field_id": field_id,
            "file": bam_file,
            "reads_mapped": stats["mapped"],
            "reads_unmapped": stats["unmapped"],
        },
        parents=[
            "children",
            {
                "id": "read_coverage",
                "datatype": "integer",
                "clamp": 1 if fields["read_cov_range"][0] == 0 else False,
                "range": fields["read_cov_range"],
            },
            "children",
        ],
    )
    return fields


def apply_filter(ids, fastq_files, **kwargs):
    """Filter FASTQ file based on read alignment file."""
    suffix = kwargs["--suffix"]
    bam_file = kwargs["--cov"]
    read_ids = set()
    filetype_letter = Path(bam_file).suffix[1]
    index_file = Path("%s.csi" % bam_file)
    if not index_file.is_file():
        pysam.index(bam_file)
    else:
        index_file = False
    samfile = pysam.AlignmentFile(bam_file, "r%s" % filetype_letter)
    for seq_id in tqdm(ids):
        for read in samfile.fetch(seq_id):
            read_ids.add(read.query_name)
    samfile.close()
    if index_file:
        os.remove(index_file)
    for fastq_file in fastq_files:
        path = Path(fastq_file)
        outfile = path.parent / (path.stem + "." + suffix + path.suffix)
        seqtk_subseq(fastq_file, "\n".join(list(read_ids)), outfile)


def parse_json_cov(json_file, **kwargs):
    """Parse coverage from JSON cov file."""
    parts = json_file.split("=")
    base_name = parts[1]
    data = load_yaml(parts[0])
    covs = []
    if "values" in data:
        for value in data["values"]:
            covs.append(float("%.4f" % value))
    if base_name.endswith("_read_cov"):
        type = "read_cov"
        parent = "read_coverage"
        datatype = "float"
        clamp = 1
    elif base_name.endswith("_cov"):
        type = "cov"
        parent = "base_coverage"
        datatype = "integer"
        clamp = 0.01
    else:
        return None
    field_id = base_name
    fields = {}
    fields["%s_id" % type] = field_id
    fields["%s_range" % type] = [
        min(covs + [kwargs["%s_range" % type][0]]),
        max(covs + [kwargs["%s_range" % type][1]]),
    ]
    if kwargs["meta"].has_field(field_id):
        file_name = kwargs["meta"].field_meta(field_id)["file"]
    else:
        file_name = json_file
    fields[type] = Variable(
        field_id,
        values=covs,
        meta={"field_id": field_id, "file": file_name},
        parents=[
            "children",
            {
                "id": parent,
                "datatype": "integer",
                "clamp": clamp if fields["%s_range" % type][0] == 0 else False,
                "range": fields["%s_range" % type],
            },
            "children",
        ],
    )
    return fields


def base_names(files):
    """Set file base names."""
    names = {}
    unique = True
    for file in files:
        if "=" in file:
            parts = file.split("=")
            name = parts[1]
            names[name] = parts[0]
        else:
            name = Path(file).stem
            if name in names:
                unique = False
            else:
                names[name] = file
    if not unique:
        print("ERROR: Unable to set unique names for coverage files")
        sys.exit(1)
    return ["%s=%s" % (v, k) for k, v in names.items()]


def parse(files, **kwargs):
    """Parse all BAM files."""
    parsed = []
    if kwargs["meta"].has_field("base_coverage"):
        cov_range = kwargs["meta"].field_meta("base_coverage")["range"]
    else:
        cov_range = [math.inf, -math.inf]
    if kwargs["meta"].has_field("read_coverage"):
        read_cov_range = kwargs["meta"].field_meta("read_coverage")["range"]
    else:
        read_cov_range = [math.inf, -math.inf]
    names = base_names(files)
    for file in names:
        if ".json" in file:
            fields = parse_json_cov(
                file, **kwargs, cov_range=cov_range, read_cov_range=read_cov_range
            )
        else:
            fields = parse_bam(
                file, **kwargs, cov_range=cov_range, read_cov_range=read_cov_range
            )
        if "cov" in fields:
            parsed.append(fields["cov"])
            cov_range = fields["cov_range"]
            if "y" not in kwargs["meta"].plot:
                kwargs["meta"].plot.update({"y": fields["cov_id"]})
        if "read_cov" in fields:
            parsed.append(fields["read_cov"])
            read_cov_range = fields["read_cov_range"]
    return parsed


def parent():
    """Set standard metadata for Coverage."""
    coverage = {
        "datatype": "float",
        "type": "variable",
        "scale": "scaleLog",
        "id": "coverage",
        "name": "coverage",
    }
    return [coverage]


def weighted_mean(values, weights, log=False):
    """Calculate weighted mean and standard deviation."""
    if log:
        mean = sum(
            [
                math.log10(value + 0.01) * weight
                for value, weight in zip(values, weights)
            ]
        ) / sum(weights)
        mean = 10 ** mean - 0.01
    else:
        mean = sum([value * weight for value, weight in zip(values, weights)]) / sum(
            weights
        )
    return mean


def summarise(indices, fields, **kwargs):
    """Summarise coverage."""
    summary = {}
    for library in fields["libraries"]:
        stats = {}
        covs = [fields["%s_cov" % library].values[i] for i in indices]
        lengths = [fields["length"].values[i] for i in indices]
        coverage = weighted_mean(covs, lengths, log=True)
        platform = kwargs["meta"].reads[library]["platform"]
        strategy = "single"
        if "strategy" in kwargs["meta"].reads[library]:
            strategy = kwargs["meta"].reads[library]["strategy"]
        elif (
            "file" in kwargs["meta"].reads[library]
            and ";" in kwargs["meta"].reads[library]["file"]
        ):
            strategy = "paired"
        stats.update(
            {
                "coverage": float("%.3f" % coverage),
                "platform": platform,
                "strategy": strategy,
            }
        )
        if "url" in kwargs["meta"].reads[library]:
            stats.update({"url": kwargs["meta"].reads[library]["url"]})
        summary.update({library: stats})
    return summary


def remove_from_meta(meta):
    """Delete all coverage fields."""
    field_ids = []
    if meta.has_field("coverage"):
        field_ids += meta.remove_field("coverage")
        meta.plot.pop("y", None)
    meta.reads = {}
    return field_ids



#!/usr/bin/env python3

# pylint: disable=too-many-locals,too-many-branches,unused-argument,unused-variable
# pylint: disable=too-many-arguments,too-many-statements

"""Parse BLAST results into MultiArray Field."""

import math
import re
from collections import Counter
from collections import defaultdict
from itertools import groupby
from operator import itemgetter

from ..lib import file_io
from .field import Category
from .field import MultiArray
from .field import Variable


def parse_blast(blast_file, cols, results=None, index=0, evalue=1, bitscore=1):
    """Parse file into dict of lists."""
    if results is None:
        results = defaultdict(list)
    bitscores = {}
    blastp = {}

    for line in file_io.stream_file(blast_file):
        row = line.rstrip().split("\t")
        score = float(row[cols["bitscore"]])
        if score < bitscore:
            continue
        if len(row) == 4:
            cols["sseqid"] = 3
        else:
            if evalue < float(row[cols["evalue"]]):
                continue
        # allow for mis-specified columns following documentation bug
        if "sstart" in cols and "qstart" not in cols:
            cols["qstart"] = cols["sstart"]
        if "send" in cols and "qend" not in cols:
            cols["qend"] = cols["send"]
        seq_id, *offset = row[cols["qseqid"]].split("_-_")
        offset = int(offset[0]) if offset else 0
        query = row[cols["qseqid"]]
        if ":" in query and "=" in query:
            # parse blastp
            parts = query.split("=")
            if query in bitscores and score <= bitscores[query]:
                continue
            if len(parts) == 3 and parts[2] == "fragmented":
                continue
            bitscores[query] = score
            seq_id, start, end = re.split(r"[:-]", parts[0])
            hit = {
                "subject": row[cols["sseqid"]],
                "score": score,
                "start": int(start),
                "end": int(end),
                "file": index,
                "title": parts[1],
            }
        else:
            # parse blastx/blastn
            try:
                hit = {
                    "subject": row[cols["sseqid"]],
                    "score": score,
                    "start": int(row[cols["qstart"]]) + offset,
                    "end": int(row[cols["qend"]]) + offset,
                    "file": index,
                }
            except IndexError:
                # parse file without positions
                hit = {
                    "subject": row[cols["sseqid"]],
                    "score": score,
                    "start": None,
                    "end": None,
                    "file": index,
                }
        try:
            taxid = row[cols["staxids"]]
            try:
                taxid, *rest = taxid.split(";")
            except ValueError:
                # no taxid for this row
                pass
            hit.update({"taxid": int(taxid)})
        except ValueError:
            # no taxid in file
            hit.update({"taxid": 0})
        if bitscores:
            blastp[query] = hit
        else:
            results[seq_id].append(hit)
    if bitscores:
        for query, hit in blastp.items():
            seq_id, rest = query.split(":")
            results[seq_id].append(hit)
    return results


def chunk_size(value):
    """Calculate nice value for chunk size."""
    mag = math.floor(math.log10(value))
    first = int(str(value)[:2]) + 1
    size = first * pow(10, mag - 1)
    return size


def set_windows(meta, taxrule):
    """Set window sizes from dataset metadata."""
    windows = []
    chunk = 0.1
    try:
        if "bestsum" in taxrule:
            chunk = 1
        elif "blast_max_chunks" in meta.settings:
            chunk = 1 / int(meta.settings["blast_max_chunks"])
        if "stats_windows" in meta.settings:
            for x in meta.settings["stats_windows"]:
                window = {
                    "key": ("%f" % x).rstrip("0").rstrip("."),
                    "window": True,
                    "value": float(x),
                    "chunk": False,
                }
                if window["key"] == "0.1":
                    window["title"] = "windows"
                else:
                    window["title"] = "windows_%s" % window["key"]
                if chunk is not None and float(x) == chunk:
                    window.update({"chunk": True})
                    chunk = None
                windows.append(window)
    except AttributeError:
        pass
    if chunk is not None:
        windows.append(
            {
                "key": ("%f" % chunk).rstrip("0").rstrip("."),
                "window": False,
                "value": chunk,
                "chunk": True,
            }
        )
    return windows


def bin_hits(
    blast,
    meta,
    identifiers,
    lengths,
    taxrule,
    hit_count,
):
    """Place hits into bins for each window size."""
    windows = set_windows(meta, taxrule)
    bins = defaultdict(dict)
    get_item = itemgetter("start")
    get_score = itemgetter("score")
    for idx, identifier in enumerate(identifiers.values):
        length = lengths[idx]
        hits = sorted(blast[identifier], key=get_item)
        for window in windows:
            if window["value"] > 1:
                chunk = window["value"]
            elif window["value"] == 1:
                chunk = length
            else:
                chunk = round(length * window["value"] / 1000 + 0.5) * 1000
            if length < 1000000:
                if not window["chunk"]:
                    continue
                chunk = 100000
            bin = {**window}
            groups = {
                k: list(g)
                for k, g in groupby(hits, lambda x: get_item(x) // chunk * chunk)
            }
            current = 0
            bin["bins"] = []
            while current < length:
                if current in groups:
                    bin["bins"].append(
                        sorted(groups[current], key=get_score, reverse=True)[:hit_count]
                    )
                else:
                    bin["bins"].append([])
                current += chunk
            bins[identifier][window["key"]] = bin
    return bins


def initialise_values(bins, prefix, results, ranks):
    """Initialise results and values ready to apply taxrule."""
    windows = [
        obj["title"]
        for obj in bins[list(bins.keys())[0]].values()
        if "title" in obj and obj["window"]
    ]
    if results is None:
        blank = [None] * len(bins.keys())
        results = []
        for rank in ranks:
            result = {
                "field_id": "%s_%s" % (prefix, rank),
                "values": blank[:],
                "data": {
                    "cindex": blank[:],
                    "score": blank[:],
                    "positions": blank[:],
                    "hits": blank[:],
                },
            }
            for window in windows:
                result["data"].update({window: blank[:]})
            results.append(result)
    values = []
    for rank in ranks:
        value = {
            "category": defaultdict(str),
            "cindex": defaultdict(int),
            "score": defaultdict(float),
            "positions": defaultdict(list),
            "categories": defaultdict(list),
            "scores": defaultdict(list),
            "hits": defaultdict(list),
        }
        for window in windows:
            value.update({window: defaultdict(list)})
        values.append(value)
    return results, values


def apply_taxrule_to_bin(seqid, values, bin, window, taxdump, ranks):
    """Apply taxrule to a single bin."""
    for index, rank in enumerate(ranks):
        cat_scores = defaultdict(float)
        for hit in bin:
            try:
                category = taxdump.ancestors[hit["taxid"]][rank]
            except KeyError:
                category = 0
            if category > 0:
                category = taxdump.names[category]
            elif category < 0:
                category = "%s-undef" % taxdump.names[-category]
            else:
                category = "undef"
            # category = "no-hit"
            if category != "undef":
                cat_scores[category] += hit["score"]
            if index == 0 and window["chunk"]:
                try:
                    values[index]["hits"][seqid].append(
                        [
                            hit["taxid"],
                            hit["start"],
                            hit["end"],
                            hit["score"],
                            hit["subject"],
                            hit["file"],
                            hit.get("title", None),
                        ]
                    )
                except KeyError:
                    pass
            if window["chunk"]:
                values[index]["positions"][seqid].append([category])
        if cat_scores:
            category = max(cat_scores, key=cat_scores.get)
            score = cat_scores[category]
        else:
            category = None
            score = 0
        if "title" in window and window["title"] in values[index]:
            values[index][window["title"]][seqid].append([category])
        if window["chunk"]:
            values[index]["categories"][seqid].append([category])
            values[index]["scores"][seqid].append(score)


def apply_taxrule_across_bins(seqid, values, window, ranks):
    """Find most common bin category."""
    if window["chunk"]:
        for index, rank in enumerate(ranks):
            scores = defaultdict(float)
            counts = defaultdict(int)
            for idx, category in enumerate(values[index]["categories"][seqid]):
                if category and category[0] is not None:
                    counts[category[0]] += 1
                    scores[category[0]] += values[index]["scores"][seqid][idx]
            max_score = 0
            max_count = 0
            top_cat = False
            for category, count in counts.items():
                if count > max_count:
                    max_count = count
                    max_score = scores[category]
                    top_cat = category
                elif count == max_count:
                    score = scores[category]
                    if score > max_score:
                        max_count = count
                        max_score = score
                        top_cat = category
            if top_cat is not False:
                values[index]["category"][seqid] = top_cat
                values[index]["cindex"][seqid] = len(counts.keys()) - 1
                values[index]["score"][seqid] = max_score


def add_values_to_results(results, values, ranks, identifiers):
    """Add new values to result set."""
    for index, rank in enumerate(ranks):
        if not identifiers.validate_list(list(values[index]["category"].keys())):
            raise UserWarning(
                "Contig names in the hits file do not match dataset identifiers."
            )
        for i, seq_id in enumerate(identifiers.values):
            if results[index]["data"]["score"][i] in (0, None):
                if seq_id in values[index]["category"]:
                    results[index]["values"][i] = values[index]["category"][seq_id]
                    results[index]["data"]["score"][i] = values[index]["score"][seq_id]
                    results[index]["data"]["cindex"][i] = values[index]["cindex"][
                        seq_id
                    ]
                    results[index]["data"]["positions"][i] = values[index]["positions"][
                        seq_id
                    ]
                    if index == 0:
                        results[index]["data"]["hits"][i] = values[index]["hits"][
                            seq_id
                        ]
                else:
                    results[index]["values"][i] = "no-hit"
                    results[index]["data"]["score"][i] = 0
                    results[index]["data"]["cindex"][i] = 0
                    results[index]["data"]["positions"][i] = []
                    if index == 0:
                        results[index]["data"]["hits"][i] = []
                for key in values[index].keys():
                    if key.startswith("windows"):
                        results[index]["data"][key][i] = values[index][key][seq_id]


def apply_taxrule(bins, taxdump, taxrule, prefix, results, identifiers):
    """Apply taxrule to binned BLAST results."""
    ranks = taxdump.list_ranks()
    # ranks = ["superkingdom"]
    results, values = initialise_values(bins, prefix, results, ranks)
    for seqid, windows in bins.items():
        for key, window in windows.items():
            for bin in window["bins"]:
                apply_taxrule_to_bin(seqid, values, bin, window, taxdump, ranks)
            apply_taxrule_across_bins(seqid, values, window, ranks)
    add_values_to_results(results, values, ranks, identifiers)
    return results


def create_fields(results, taxrule, files, fields=None):
    """Store BLAST results as Fields."""
    if fields is None:
        fields = []
    hits_id = "%s_%s" % (taxrule, "positions")
    fields.append(
        MultiArray(
            hits_id,
            values=results[0]["data"]["hits"],
            meta={
                "field_id": hits_id,
                "name": hits_id,
                "type": "multiarray",
                "datatype": "mixed",
                "preload": False,
                "active": False,
                "files": files,
            },
            parents=["children", {"id": taxrule}, "children"],
            category_slot=None,
            headers=["taxid", "start", "end", "score", "subject", "index", "title"],
        )
    )
    for result in results:
        main = Category(
            result["field_id"],
            values=result["values"],
            meta={"field_id": result["field_id"], "name": result["field_id"]},
            parents=["children", {"id": taxrule}, "children"],
        )
        fields.append(main)
        parents = [
            "children",
            {"id": taxrule},
            "children",
            {"id": result["field_id"]},
            "data",
        ]
        field_id = "%s_%s" % (result["field_id"], "cindex")
        fields.append(
            Variable(
                field_id,
                values=result["data"]["cindex"],
                meta={
                    "scale": "scaleLinear",
                    "field_id": field_id,
                    "name": field_id,
                    "datatype": "integer",
                    "range": [
                        min(result["data"]["cindex"]),
                        max(result["data"]["cindex"]),
                    ],
                    "preload": False,
                    "active": False,
                },
                parents=parents,
            )
        )
        field_id = "%s_%s" % (result["field_id"], "score")
        _min = min(result["data"]["score"])
        fields.append(
            Variable(
                field_id,
                values=result["data"]["score"],
                meta={
                    "scale": "scaleLog",
                    "field_id": field_id,
                    "name": field_id,
                    "clamp": 1 if _min == 0 else False,
                    "datatype": "float",
                    "range": [_min, max(result["data"]["score"])],
                    "preload": False,
                    "active": False,
                },
                parents=parents,
            )
        )
        subfield = "positions"
        field_id = "%s_%s" % (result["field_id"], subfield)
        if len(result["data"][subfield]) > 1:
            headers = ["name"]
        else:
            headers = ["name"]
        fields.append(
            MultiArray(
                field_id,
                values=result["data"][subfield],
                fixed_keys=main.keys,
                meta={
                    "field_id": field_id,
                    "name": field_id,
                    "type": "multiarray",
                    "datatype": "string",
                    "preload": False,
                    "active": False,
                    "linked_field": hits_id,
                },
                parents=parents,
                category_slot=0,
                headers=headers,
            )
        )
        for subfield in result["data"].keys():
            if subfield.startswith("windows"):
                field_id = "%s_%s" % (result["field_id"], subfield)
                if len(result["data"][subfield]) > 1:
                    headers = ["name"]
                else:
                    headers = ["name"]
                fields.append(
                    MultiArray(
                        field_id,
                        values=result["data"][subfield],
                        fixed_keys=main.keys,
                        meta={
                            "field_id": field_id,
                            "name": field_id,
                            "type": "array",
                            "datatype": "string",
                            "preload": False,
                            "active": False,
                        },
                        parents=parents,
                        category_slot=0,
                        headers=headers,
                    )
                )

    return fields


def parse(files, **kwargs):
    """Parse BLAST results into Fields."""
    blast = None
    fields = []
    identifiers = kwargs["dependencies"]["identifiers"]
    lengths = kwargs["dependencies"]["length"].values
    try:
        taxrule, prefix = kwargs["--taxrule"].split("=")
    except ValueError:
        taxrule = kwargs["--taxrule"]
        prefix = taxrule
    cols = {}
    columns = kwargs["--hits-cols"].split(",")
    for column in columns:
        try:
            index, name = column.split("=")
            cols[name] = int(index) - 1
        except ValueError:
            exit("ERROR: --hits-cols contains an invalid value.")
    if taxrule.endswith("order"):
        results = None
        for index, file in enumerate(files):
            blast = parse_blast(
                file,
                cols,
                None,
                index,
                float(kwargs["--evalue"]),
                float(kwargs["--bitscore"]),
            )
            bins = bin_hits(
                blast,
                kwargs["meta"],
                identifiers,
                lengths,
                taxrule,
                int(kwargs["--hit-count"]),
            )
            results = apply_taxrule(
                bins, kwargs["taxdump"], taxrule, prefix, results, identifiers
            )
        fields = create_fields(results, prefix, files)
    else:
        results = None
        previous = None
        for index, file in enumerate(files):
            blast = parse_blast(
                file,
                cols,
                previous,
                index,
                float(kwargs["--evalue"]),
                float(kwargs["--bitscore"]),
            )
            previous = blast
        bins = bin_hits(
            blast,
            kwargs["meta"],
            identifiers,
            lengths,
            taxrule,
            int(kwargs["--hit-count"]),
        )
        results = apply_taxrule(
            bins, kwargs["taxdump"], taxrule, prefix, results, identifiers
        )
        fields = create_fields(results, prefix, files)
    update_plot = kwargs.get("--update-plot", False)
    if update_plot or "cat" not in kwargs["meta"].plot:
        kwargs["meta"].plot.update({"cat": "%s_phylum" % prefix})
    return fields


def parent():
    """Set standard metadata for BLAST."""
    blast = {
        "datatype": "string",
        "type": "category",
        "id": "taxonomy",
        "name": "Taxonomy",
    }
    return [blast]


def summarise(indices, fields, **kwargs):
    """Summarise assembly sequence stats."""
    taxonomy = kwargs["stats"]["taxonomy"]
    summary = {}
    lengths = [fields["length"].values[i] for i in indices]
    gcs = [fields["gc"].values[i] for i in indices]
    covs = [fields["cov"].values[i] for i in indices] if "cov" in fields else []
    summary.update({"total": length_stats(lengths, gcs, covs)})
    hits = list(
        map(
            lambda x: fields["hits"].keys[x],
            [fields["hits"].values[i] for i in indices],
        )
    )
    counts = Counter(hits)
    index = 0
    limit = 9
    other = []
    other_gcs = []
    other_covs = []
    for key in sorted(counts.items(), key=lambda x: x[1], reverse=True):
        taxon = key[0]
        subset = [i for i, j in enumerate(hits) if j == taxon]
        lengths = [fields["length"].values[indices[i]] for i in subset]
        gcs = [fields["gc"].values[indices[i]] for i in subset]
        covs = (
            [fields["cov"].values[indices[i]] for i in subset]
            if "cov" in fields
            else []
        )
        if len(counts.keys()) > limit and index >= limit:
            other += lengths
            other_gcs += gcs
            other_covs += covs
            if "target" in taxonomy and taxon == taxonomy["target"]:
                summary.update({"target": length_stats(lengths, gcs, covs)})
        else:
            summary.update({taxon: length_stats(lengths, gcs, covs)})
        index += 1
    if other:
        summary.update({"other": length_stats(other, other_gcs, other_covs)})
    return summary


def weighted_mean_sd(values, weights, log=False):
    """Calculate weighted mean, median and standard deviation."""
    weighted = []
    total_weight = 0
    total = 0
    for weight, value in zip(weights, values):
        if log:
            value = math.log10(value) + 10
        weighted.append({"weight": weight, "value": value})
        total_weight += weight
        total += value * weight
    mid = total_weight / 2
    running_total = 0
    median = 0
    for entry in sorted(weighted, key=lambda i: i["value"]):
        running_total += entry["weight"]
        if running_total > mid:
            median = entry["value"]
            break
    mean = total / total_weight
    if median == 0:
        median = mean
    variance = (
        sum([entry["weight"] * ((entry["value"] - mean) ** 2) for entry in weighted])
        / total_weight
    )
    st_dev = variance ** 0.5
    upper = mean + 2 * st_dev
    lower = mean - 2 * st_dev
    if log:
        upper = 10 ** (upper - 10)
        lower = 10 ** (lower - 10)
        mean = 10 ** (mean - 10)
        median = 10 ** (median - 10)
    return mean, median, st_dev, upper, lower


def length_stats(all_lengths, all_gcs, all_covs):
    """Calculate stats for a set of sequence lengths."""
    span = sum(all_lengths)
    count = len(all_lengths)
    lengths = []
    gcs = []
    covs = []
    if all_covs:
        for length, gc_value, cov in zip(all_lengths, all_gcs, all_covs):
            if cov >= 0.01:
                lengths.append(length)
                gcs.append(gc_value)
                covs.append(cov)
    else:
        lengths = all_lengths
        gcs = all_gcs
    stats = {"span": span, "count": count}
    if gcs:
        gc_mean, gc_median, gc_dev, gc_upper, gc_lower = weighted_mean_sd(gcs, lengths)
        stats.update(
            {
                "gc": [
                    float("%.4f" % gc_mean),
                    float("%.4f" % gc_median),
                    float("%.4f" % gc_lower),
                    float("%.4f" % gc_upper),
                    float("%.4f" % min(gcs)),
                    float("%.4f" % max(gcs)),
                ]
            }
        )
    if covs:
        cov_mean, cov_median, cov_dev, cov_upper, cov_lower = weighted_mean_sd(
            covs, lengths, log=True
        )
        stats.update(
            {
                "cov": [
                    float("%.4f" % cov_mean),
                    float("%.4f" % cov_median),
                    float("%.4f" % cov_lower),
                    float("%.4f" % cov_upper),
                    float("%.4f" % min(covs)),
                    float("%.4f" % max(covs)),
                ]
            }
        )
    n50 = span * 0.5
    n90 = span * 0.9
    all_lengths.sort(reverse=True)
    nlength = 0
    ncount = 0
    for length in all_lengths:
        ncount += 1
        nlength += length
        if "n50" not in stats and nlength > n50:
            stats.update({"n50": length, "l50": ncount})
        if "n90" not in stats and nlength > n90:
            stats.update({"n90": length, "l90": ncount})
    if "n50" not in stats:
        stats.update({"n50": all_lengths[-1], "l50": ncount})
    if "n90" not in stats:
        stats.update({"n90": all_lengths[-1], "l90": ncount})
    return stats


def remove_from_meta(meta):
    """Delete all hits fields."""
    field_ids = []
    if meta.has_field("taxonomy"):
        field_ids += meta.remove_field("taxonomy")
        meta.plot.pop("cat", None)
    return field_ids



#!/usr/bin/env python3

# pylint: disable=no-member, too-many-branches, too-many-nested-blocks

"""
Field-based calculations.

Usage:
    blobtools calc [--combine STRING...] [--equation STRING...] [--sum LIST...]
                   [--replace] DIRECTORY

Arguments:
    DIRECTORY             Existing Blob directory.

Options:
    --combine   STRING    Group categories FIELD:CATEGORY,LIST:ALIAS.
    --equation  STRING    Combine variable fields using ()/*+- operators, e.g. base_cov*(length-ncount).
    --sum       LIST      Sum all fields (and/or child fields) in a comman separated list of field IDs.
    --replace             Replace existing fields with matching ids.

Examples:
    blobtools calc --sum base_coverage=summed_cov
    blobtools calc --equation ncount/length=n_fraction
    blobtools calc --equation ncount/length*100=n_percent
    blobtools calc --equation (base_cov1+base_cov2)*(length-ncount)=combined_base_count
    blobtools calc --combine bestdistorder_species:(bestdistorder_phylum:Nematoda):Nematoda=nematoda_grouped
    blobtools calc --combine bestdistorder_species:(bestdistorder_phylum:Nematoda,Arthropoda):Contaminants=contaminants
    blobtools calc --combine bestdistorder_genus:Homo,Pan:Apes=apes_grouped
    blobtools calc --combine bestdistorder_genus:Homo,Pan:Apes::bestdistorder_genus:Mus,Rattus:Rodents=apes_and_rodents
    blobtools calc --combine bestdistorder_genus:(bestdistorder_genus_cindex:>0):Uncertain=uncertain_grouped

"""

from docopt import docopt

# from fetch import fetch_metadata


def has_field_warning(meta, field_id):
    """Warn if dataset has existing field with same id."""
    if meta.has_field(field_id):
        print(
            "WARN: Field '%s' is already present in dataset, not overwriting."
            % field_id
        )
        print("WARN: Use '--replace' flag to overwrite existing field.")
        return 1
    return 0


def main(args):
    """Entrypoint for blobtools calc."""
    # meta = fetch_metadata(args['DIRECTORY'], **args)
    if args["--sum"]:
        print(args["--sum"])
    if args["--equation"]:
        print(args["--equation"])
    if args["--combine"]:
        print(args["--combine"])

    # if not args['--replace']:
    #     if has_field_warning(meta, data.field_id):
    #         continue
    #     for parent in data.parents:
    #         if 'range' in parent:
    #             parent_meta = meta.field_meta(parent['id'])
    #             if parent_meta and 'range' in parent_meta:
    #                 parent['range'][0] = min(parent['range'][0], parent_meta['range'][0])
    #                 parent['range'][1] = max(parent['range'][1], parent_meta['range'][1])
    #     meta.add_field(parents+data.parents, **data.meta)
    #     json_file = "%s/%s.json" % (args['DIRECTORY'], data.field_id)
    #     file_io.write_file(json_file, data.values_to_dict())
    # file_io.write_file("%s/meta.json" % args['DIRECTORY'], meta.to_dict())


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

# pylint: disable=no-member, too-many-branches, too-many-locals, too-many-statements, broad-except

"""
Generate plots using BlobToolKit Viewer.

Usage:
  blobtools view [--format STRING...] [--host STRING] [--interactive]
  [--out PATH] [--param STRING...] [--ports RANGE] [--prefix STRING]
  [--preview STRING...] [--driver STRING] [--driver-log PATH]
  [--local] [--remote] [--timeout INT] [--view STRING...] DIRECTORY

Options:
      --format STRING         Image format (svg|png). [Default: png]
      --host STRING           Hostname. [Default: http://localhost]
      --interactive           Start interactive session (opens dataset in Firefox/Chromium). [Default: False]
      --out PATH              Directory for outfiles. [Default: .]
      --param key=value       Query string parameter.
      --ports RANGE           Port range for viewer and API. [Default: 8000-8099]
      --prefix STRING         URL prefix. [Default: view]
      --preview STRING        Field name.
      --driver STRING         Webdriver to use (chromium or firefox). [Default: firefox]
      --driver-log PATH       Path to driver logfile for debugging. [Default: /dev/null]
      --local                 Start viewer for local session. [Default: False]
      --remote                Start viewer for remote session. [Default: False]
      --timeout INT           Time to wait for page load in seconds. Default (0) is no timeout. [Default: 0]
      --view STRING           Plot type (blob|cumulative|snail). [Default: blob]
"""
import os
import shlex
import signal
import sys
import time
from pathlib import Path
from shutil import which
from subprocess import PIPE
from subprocess import Popen
from traceback import format_exc

import psutil
from docopt import docopt
from pyvirtualdisplay import Display
from selenium import webdriver
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.common.by import By
from selenium.webdriver.firefox.options import Options
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait
from tolkein import tolog

from .host import test_port
from .version import __version__

LOGGER = tolog.logger(__name__)


def file_ready(file_path, timeout, callback):
    """Check if file is ready."""
    start_time = time.time()
    while not os.path.exists(file_path):
        elapsed_time = time.time() - start_time
        if timeout and elapsed_time > timeout:
            callback("Timeout waiting for file")
            return False
        # flush nfs cache by chowning parent to current owner
        parent = os.path.dirname(os.path.abspath(file_path))
        os.chown(parent, os.stat(parent).st_uid, os.stat(parent).st_gid)
        time.sleep(1)
    if os.path.isfile(file_path):
        return True
    callback("%s is not a file" % file_path)
    return False


def test_loc(args):
    """See if dataset needs to be hosted and, if so, find an empty port."""
    info = args["--host"].split(":")
    dataset = Path(args["DIRECTORY"]).name
    level = "dataset"
    if len(info) >= 2 and info[1] != "//localhost":
        loc = "%s/%s/%s/dataset/%s" % (
            args["--host"],
            args["--prefix"],
            dataset,
            dataset,
        )
        return loc, None, None, None, level
    if len(info) == 1 and info[0] != "localhost":
        # need to add test for http vs https
        loc = "http://%s/%s/%s/dataset/%s" % (
            args["--host"],
            args["--prefix"],
            dataset,
            dataset,
        )
        return loc, None, None, None, level
    if len(info) == 3:
        port = info[2]
        for i in range(0, 10):
            available = test_port(port, "test")
            if available:
                if i == 9:
                    print("ERROR: No service running on port %s" % port)
                    print("       Unable to connect to %s" % args["--host"])
                    sys.exit(1)
                time.sleep(1)
            else:
                loc = "%s/%s/%s/dataset/%s" % (
                    args["--host"],
                    args["--prefix"],
                    dataset,
                    dataset,
                )
                return loc, None, None, None, level
    if args["DIRECTORY"] == "_":
        parent = "_"
        level = "blobdir"
    else:
        if not Path(args["DIRECTORY"]).exists():
            print(
                "ERROR: DIRECTORY '%s' must be a valid path to begin hosting."
                % args["DIRECTORY"]
            )
            sys.exit(1)
        dataset = Path(args["DIRECTORY"]).name
        if (
            Path("%s/meta.json" % args["DIRECTORY"]).is_file()
            or Path("%s/meta.json.gz" % args["DIRECTORY"]).is_file()
        ):
            parent = Path(args["DIRECTORY"]).resolve().absolute().parent
        else:
            level = "blobdir"
            parent = Path(args["DIRECTORY"]).resolve().absolute()
    port_range = args["--ports"].split("-")
    api_port = False
    port = False
    for i in range(int(port_range[0]), int(port_range[1])):
        if test_port(i, "test"):
            if not api_port:
                api_port = i
                continue
            if not port:
                port = i
            break
    # directory = Path(__file__).resolve().parent.parent
    cmd = "blobtools host --port %d --api-port %d %s" % (port, api_port, parent)
    process = Popen(shlex.split(cmd), stdout=PIPE, stderr=PIPE, encoding="ascii")
    loc = "%s:%d/%s" % (args["--host"], port, args["--prefix"])
    if level == "dataset":
        loc += "/%s/dataset/%s" % (dataset, dataset)
    else:
        loc += "/all"
    print("Initializing viewer")
    for i in range(0, 10):
        poll = process.poll()
        if poll is None:
            if test_port(port, "test"):
                break
            time.sleep(1)
        else:
            print(process.stdout.read(), file=sys.stderr)
            print(process.stderr.read(), file=sys.stderr)
            print("ERROR: Viewer quit unexpectedly", file=sys.stderr)
            print("Unable to run: %s" % cmd, file=sys.stderr)
            sys.exit(1)
    return loc, process, port, api_port, level


def firefox_driver(args):
    """Start firefox."""
    if not (which("firefox") or which("geckodriver")):
        try:
            LOGGER.info("Firefox and geckodriver must be available for blobtools view")
            LOGGER.info("Attempting to install the appropriate geckodriver version")
            import geckodriver_autoinstaller

            geckodriver_autoinstaller.install()
            LOGGER.info("Successfully installed geckodriver")

        except Exception:
            LOGGER.error(
                "Unable to install automatically. Try `conda install -c conda-forge firefox geckodriver` or `blobtools view --driver chromium` to use chrome browser."
            )
            sys.exit(1)

    outdir = os.path.abspath(args["--out"])
    os.makedirs(Path(outdir), exist_ok=True)

    profile = webdriver.FirefoxProfile()
    profile.set_preference("browser.download.folderList", 2)
    profile.set_preference("browser.download.manager.showWhenStarting", False)
    profile.set_preference("browser.download.dir", outdir)
    profile.set_preference("browser.download.lastDir", args["--out"])
    profile.set_preference(
        "browser.helperApps.neverAsk.saveToDisk",
        "image/png, image/svg+xml, text/csv, text/plain, application/json",
    )

    options = Options()
    options.headless = not args["--interactive"]
    display = Display(visible=0, size=(800, 600))
    display.start()

    driver = webdriver.Firefox(
        options=options, firefox_profile=profile, service_log_path=args["--driver-log"],
    )
    time.sleep(2)
    return driver, display


def chromium_driver(args):
    """Start chromium browser."""
    os.environ["WDM_LOG_LEVEL"] = "0"
    try:
        LOGGER.info("Chrome and chromedriver must be available for blobtools view")
        LOGGER.info("Attempting to install the appropriate chromedriver version")
        from webdriver_manager.chrome import ChromeDriverManager

        service_object = Service(ChromeDriverManager().install())
        LOGGER.info("Successfully installed chromedriver")
    except Exception:
        LOGGER.error(
            "Unable to install automatically. Check Chrome is available or try `blobtools view --driver firefox` to use firefox browser."
        )
        sys.exit(1)
    outdir = os.path.abspath(args["--out"])
    os.makedirs(Path(outdir), exist_ok=True)

    options = webdriver.ChromeOptions()
    if not args["--interactive"]:
        options.add_argument("headless")
    prefs = {}
    prefs["profile.default_content_settings.popups"] = 0
    prefs["download.default_directory"] = outdir
    options.add_experimental_option("prefs", prefs)
    display = Display(visible=0, size=(800, 600))
    display.start()
    driver = webdriver.Chrome(
        options=options,
        service=service_object,
        # executable_path=add option to set binary location,
        service_log_path=args["--driver-log"],
    )
    time.sleep(2)
    return driver, display


def static_view(args, loc, viewer):
    """Generate static images."""
    qstr = "staticThreshold=Infinity"
    qstr += "&nohitThreshold=Infinity"
    qstr += "&plotGraphics=svg"
    file_stem = Path(args["DIRECTORY"]).name
    if file_stem == "_":
        file_stem = "FXWY01"
    if args["--format"] == "svg":
        qstr += "&svgThreshold=Infinity"
    shape = "circle"
    for param in args["--param"]:
        qstr += "&%s" % str(param)
        key, value = param.split("=")
        if key == "plotShape":
            shape = value

    timeout = int(args["--timeout"])
    outdir = os.path.abspath(args["--out"])
    driver = None

    def handle_error(err):
        """Release resources before quitting."""
        if viewer is not None:
            viewer.send_signal(signal.SIGINT)
        if driver is not None:
            driver.quit()
            display.stop()
        print(err)
        sys.exit(1)

    if args["--driver"] == "firefox":
        """View dataset in Firefox."""
        try:
            driver, display = firefox_driver(args)
        except Exception as err:
            format_exc(err)
            LOGGER.error(
                "Unable to start Firefox. Try `conda install -c conda-forge firefox geckodriver` or `blobtools view --driver chromium` to use Chrome browser."
            )
            sys.exit(1)

    elif args["--driver"] == "chromium":
        """View dataset in Chromium browser."""
        try:
            driver, display = chromium_driver(args)
        except Exception as err:
            LOGGER.error(
                "Unable to start Chrome. Try `blobtools view --driver firefox` to use Firefox browser."
            )
            sys.exit(1)
    else:
        handle_error("%s is not a valid driver" % args["--driver"])

    try:
        view = args["--view"][0]
        if args["--preview"]:
            qstr += "#Filters"
        url = "%s/%s?%s" % (loc, view, qstr)
        print("Loading %s" % url)
        for i in range(0, 10):
            try:
                driver.get(url)
            except Exception as err:
                if i < 5:
                    time.sleep(1)
                    continue
                handle_error(err)
            break

        for next_view in args["--view"]:
            if next_view != view:
                view = next_view
                url = "%s/%s?%s" % (loc, view, qstr)
                print("Navigating to %s" % url)
                try:
                    driver.get(url)
                except Exception as err:
                    handle_error(err)
            for fmt in args["--format"]:
                file = "%s.%s" % (file_stem, view)
                if view == "blob":
                    file += ".%s" % shape
                elif view == "busco":
                    view = "all_%s" % view
                    if fmt not in ("csv", "json"):
                        fmt = "json"
                file += ".%s" % fmt
                print("Fetching %s" % file)
                el_id = "%s_save_%s" % (view, fmt)
                print("waiting for element %s" % el_id)
                unstable = True
                start_time = time.time()
                while unstable:
                    elapsed_time = time.time() - start_time
                    if timeout and elapsed_time > timeout:
                        handle_error("Timeout waiting for file")
                    try:
                        element = WebDriverWait(driver, timeout).until(
                            EC.visibility_of_element_located((By.ID, el_id))
                        )
                        element.click()
                        unstable = False
                        file_name = "%s/%s" % (outdir, file)
                        print("waiting for file '%s'" % file_name)
                        file_ready(file_name, timeout, handle_error)
                    except Exception as err:
                        unstable = True
                        time.sleep(1)

        for preview in args["--preview"]:
            print("Creating %s preview" % preview)
            for fmt in args["--format"]:
                el_id = "%s_preview_save_%s" % (preview, fmt)
                file = "%s.%s.preview.%s" % (Path(args["DIRECTORY"]).name, preview, fmt)
                try:
                    element = WebDriverWait(driver, timeout).until(
                        EC.visibility_of_element_located((By.ID, el_id))
                    )
                    element.click()
                    file_name = "%s/%s" % (outdir, file)
                    print("waiting for file '%s'" % file_name)
                    file_ready(file_name, timeout, handle_error)
                except Exception as err:
                    handle_error(err)
        if viewer is not None:
            viewer.send_signal(signal.SIGINT)
        driver.quit()
        display.stop()
    except Exception as err:
        handle_error(err)
        # print(err)
        # if viewer is not None:
        #     viewer.send_signal(signal.SIGINT)
        # driver.quit()
        # display.stop()
    return True


def interactive_view(args, loc, viewer, level):
    if args["--driver"] == "firefox":
        """View dataset in Firefox."""
        driver, display = firefox_driver(args)
    elif args["--driver"] == "chromium":
        """View dataset in Chromium browser."""
        driver, display = chromium_driver(args)
    qstr = ""
    for param in args["--param"]:
        qstr += "&%s" % str(param)
    try:
        view = args["--view"][0]
        if args["--preview"]:
            qstr += "#Filters"
        if level == "dataset":
            url = "%s/%s" % (loc, view)
            if qstr:
                url += "?%s" % qstr
        else:
            url = loc if loc.endswith("all") else "%s/all" % loc
        print("Loading %s" % url)
        try:
            driver.get(url)
        except Exception as err:
            print(err)
        poll = viewer.poll()
        while poll is None:
            time.sleep(5)
            poll = viewer.poll()
        driver.quit()
        display.stop()
        if viewer is not None:
            viewer.send_signal(signal.SIGINT)
    except Exception as err:
        print(err)
        driver.quit()
        display.stop()
        if viewer is not None:
            viewer.send_signal(signal.SIGINT)
    return True


def remote_view(args, loc, viewer, port, api_port, level, remote):
    """View dataset remotely."""
    qstr = ""
    for param in args["--param"]:
        qstr += "&%s" % str(param)
    try:
        view = args["--view"][0]
        if args["--preview"]:
            qstr += "#Filters"
        if level == "dataset":
            url = "%s/%s" % (loc, view)
            if qstr:
                url += "?%s" % qstr
            print("View dataset at %s" % url)
        else:
            print("View datasets at %s" % loc)
        if remote:
            print("For remote access use:")
            print(
                "    ssh -L %d:127.0.0.1:%d -L %d:127.0.0.1:%d username@remote_host"
                % (port, port, api_port, api_port)
            )
        while True:
            time.sleep(5)
    except Exception as err:
        print("remote exception")
        print(err)
        if viewer is not None:
            viewer.send_signal(signal.SIGINT)
    return True


def main(args):
    """Entrypoint for blobtools view."""
    loc, viewer, port, api_port, level = test_loc(args)
    try:
        if args["--interactive"]:
            interactive_view(args, loc, viewer, level)
        elif args["--remote"]:
            remote_view(args, loc, viewer, port, api_port, level, True)
        elif args["--local"]:
            remote_view(args, loc, viewer, port, api_port, level, False)
        else:
            static_view(args, loc, viewer)
    except KeyboardInterrupt:
        pass
    time.sleep(1)
    if viewer is not None:
        viewer.send_signal(signal.SIGINT)
        time.sleep(1)


def cli():
    """Entry point."""
    if len(sys.argv) == sys.argv.index(__name__.split(".")[-1]) + 1:
        args = docopt(__doc__, argv=[])
    else:
        args = docopt(__doc__, version=__version__)
    if not os.path.exists(os.environ["HOME"]):
        os.mkdir(os.environ["HOME"])
    main(args)


if __name__ == "__main__":
    cli()



#!/usr/bin/env python3
"""Functions to run external programs."""

import gzip
import shlex
from pathlib import Path
from subprocess import PIPE, run


def seqtk_subseq(infile, ids, outfile):
    """Run seqtk subseq."""
    cmd = "seqtk subseq %s -" % infile
    file_open = open
    if Path(infile).suffix == '.gz':
        file_open = gzip.open
    with file_open(outfile, 'wt') as ofh:
        process = run(shlex.split(cmd),
                      stdout=PIPE,
                      stderr=PIPE,
                      input=ids,
                      encoding='ascii')
        ofh.write(process.stdout)
    return process.returncode


if __name__ == '__main__':
    import doctest
    doctest.testmod()



#!/usr/bin/env python3
"""Parse BUSCO results into MultiArray Field."""

import re
from pathlib import Path

from ..lib import file_io
from .field import Array
from .text import parse_header_row
from .text import set_delimiter


def parse_synonyms(synonym_file, delimiter, columns, header, identifiers):
    """Parse synonyms into Array."""
    meta = {}
    synonym_file, *prefix = synonym_file.split("=")
    if prefix:
        prefix = prefix[0]
    else:
        prefix = Path(synonym_file).stem
    meta["field_id"] = "%s_synonyms" % prefix
    by_id = {}
    ids = identifiers.to_set()
    data = file_io.stream_file(synonym_file)
    lines = [line for line in data]
    if columns:
        columns = columns.split(",")
    else:
        columns = []
    delimit = set_delimiter(delimiter, sample=lines[0])
    if header:
        header_row = lines[0].rstrip().replace('"', "")
        columns = parse_header_row(delimit, header_row, columns)
        lines = lines[1:]
    try:
        id_col = columns.index("identifier")
    except ValueError:
        id_col = None
    for line in lines:
        row = re.split(delimit, line.rstrip().replace('"', ""))
        key = None
        names = []
        for i, value in enumerate(row):
            if id_col is None and value in ids:
                key = value
                id_col = i
            elif i == id_col:
                key = value
            else:
                names.append(value)
        by_id.update({key: names})
    values = [by_id[id] if id in by_id else [] for id in identifiers.values]
    del columns[id_col]
    synonyms_field = Array(
        meta["field_id"],
        meta=meta,
        values=values,
        headers=columns,
        parents=["children"],
    )
    return synonyms_field


def parse(files, **kwargs):
    """Parse all synonym files."""
    parsed = []
    for file in files:
        parsed.append(
            parse_synonyms(
                file,
                delimiter=kwargs["--text-delimiter"],
                columns=kwargs["--text-cols"],
                header=kwargs["--text-header"],
                identifiers=kwargs["dependencies"]["identifiers"],
            )
        )
    return parsed


def parent():
    """Set standard metadata for synonyms."""
    synonyms = {
        "datatype": "string",
        "type": "array",
        "id": "synonyms",
        "name": "Synonyms",
    }
    return [synonyms]



#!/usr/bin/env python3

# pylint: disable=unused-argument, broad-except

"""Add taxonomic ranks to metadata for a taxid."""


def add(taxid, taxdump, meta):
    """Add ranks to meta."""
    taxid = int(taxid)
    lineage = taxdump.lineage(taxid)
    if lineage:
        meta.taxon.update({'taxid': taxid})
        for rank, value in lineage.items():
            meta.taxon.update({rank: value})
    else:
        print("WARN: '%s' was not found in the taxdump." % taxid)
    return meta.taxon


def summarise(indices, fields, **kwargs):
    """Summarise taxonomy."""
    summary = {}
    meta = kwargs['meta']
    try:
        if 'taxid' in meta.taxon:
            summary.update({'taxid': meta.taxon['taxid']})
        names = []
        for rank in ('superkingdom', 'kingdom', 'phylum', 'class', 'order', 'family', 'genus', 'species'):
            if rank in meta.taxon:
                names.append(meta.taxon[rank])
        if names:
            summary.update({'lineage': '; '.join(names)})
        if 'cat' in meta.plot:
            rank = meta.plot['cat'].split('_')[-1]
            summary.update({'targetRank': rank})
            summary.update({'target': meta.taxon[rank]})
    except Exception:
        pass
    return summary
